"""Phonon band-structure for silicon using the Appelbaum-Hamann PP.

For comparison, see e.g.:

    * Phys. Rev. B 43, 7231 (1991).

"""

import numpy as np
import matplotlib.pyplot as plt
import ase.units as units
from ase.dft.kpoints import bandpath

from gpaw.mpi import rank, world
from gpaw.dfpt import PhononCalculator

# Pseudo-potential
PP = 'AH'

# Name of file with ground-state calculation
name = 'Si_%s.gpw' % PP

# Create phonon calculator
ph = PhononCalculator(name,
                      gamma=False,
                      symmetry=False,
                      e_ph=False)

# Run the self-consistent calculation
ph.run()

# Ensure that the master does not enter here before all files have been created
world.barrier()

# Calculate band-structure and plot on master
if rank == 0:

    # High-symmetry points in the Brillouin zone
    atoms = ph.get_atoms()
    path_kc, q, Q = bandpath('GKXGLXWL', atoms.cell, 100)
    point_names = [r'$\Gamma$', 'K', 'X', r'$\Gamma$', 'L', 'X', 'W', 'L']

    # Calculate band-structure
    omega_kn = ph.band_structure(path_kc)

    # Convert from sqrt(Ha / Bohr**2 / amu) -> meV
    s = units.Hartree**0.5 * units._hbar * 1.e10 / \
        (units._e * units._amu)**(0.5) / units.Bohr
    omega_kn *= s * 1000

    # Plot the band-structure
    plt.figure(1)
    for n in range(len(omega_kn[0])):
        plt.plot(q, omega_kn[:, n], 'k-', lw=2)

    plt.xticks(Q, point_names, fontsize=18)
    plt.yticks(fontsize=18)
    plt.xlim(q[0], q[-1])
    plt.ylim(0, np.ceil(omega_kn.max() / 10) * 10)
    plt.ylabel(r'Frequency ($\mathrm{meV}$)', fontsize=22)
    plt.grid('on')
    # plt.show()
    plt.savefig('Si_bandstructure.png')



import os
import gpaw.wannier90 as w90
from gpaw import GPAW

seed = 'GaAs'

calc = GPAW(seed + '.gpw', txt=None)

w90.write_input(calc, orbitals_ai=[[], [0, 1, 2, 3]],
                bands=range(4),
                seed=seed,
                num_iter=1000,
                plot=True)
w90.write_wavefunctions(calc, seed=seed)
os.system('wannier90.x -pp ' + seed)

w90.write_projections(calc, orbitals_ai=[[], [0, 1, 2, 3]], seed=seed)
w90.write_eigenvalues(calc, seed=seed)
w90.write_overlaps(calc, seed=seed)

os.system('wannier90.x ' + seed)



# web-page: Co_magnon_dispersion.png
"""Plot the magnon dispersion of Co(hcp)"""

# General modules
import numpy as np
from matplotlib import rcParams
import matplotlib.pyplot as plt

# Script modules
from gpaw import GPAW
from gpaw.response.heisenberg import (calculate_fm_magnon_energies,
                                      calculate_single_site_magnon_energies)


# ---------- Inputs ---------- #

# Ground state
gpw = 'Co.gpw'

# High-symmetry points
sp_p = [r'$\Gamma$', 'M', 'K', 'A']
spq_pc = [np.array(qc) for qc in
          [[0., 0., 0], [0.5, 0., 0.], [1 / 3., 1 / 3., 0.], [0., 0., 0.5]]]

# Load MFT data
q_qc = np.load('Co_q_qc.npy')
rc_r = np.load('Co_rc_r.npy')
Jsph1_qabr = np.load('Co_Jsph1_qabr.npy')
Jsph2_qabr = np.load('Co_Jsph2_qabr.npy')
Jmix_qp = np.load('Co_Jmix_qp.npy')

# Define range of spherical radii to plot
rmin = 1.0
rmax = 1.4

# Define a radius to plot the inequivalent spherical site model for
rineq = 0.6

# Labels and limits
mwlabel = r'$\hbar\omega$ [meV]'
mwlim = (0., 800.)

filename = 'Co_magnon_dispersion.png'


# ---------- Script ---------- #

# Extract the magnetization of the unit cell
calc = GPAW(gpw, txt=None)
muc = calc.get_magnetic_moment()

# Convert relative q-points into distance along the bandpath in reciprocal
# space.
B_cv = 2.0 * np.pi * calc.atoms.cell.reciprocal()  # Coordinate transform
q_qv = q_qc @ B_cv  # Transform into absolute reciprocal coordinates
pathq_q = [0.]
for q in range(1, len(q_qc)):
    pathq_q.append(pathq_q[-1] + np.linalg.norm(q_qc[q] - q_qc[q - 1]))
pathq_q = np.array(pathq_q)

# Define q-limits of plot
qlim = ((pathq_q[0] - pathq_q[1]) / 2.,
        (1.5 * pathq_q[-1] - 0.5 * pathq_q[-2]))

# Calculate the magnon energies
# We distribute the unit cell magnetization evenly between the sites
mm_ar = muc / 2. * np.ones(Jsph1_qabr.shape[2:], dtype=float)
Esph1_qnr = calculate_fm_magnon_energies(Jsph1_qabr, q_qc, mm_ar) * 1.e3  # meV
Esph2_qnr = calculate_fm_magnon_energies(Jsph2_qabr, q_qc, mm_ar) * 1.e3
Emix_qp = calculate_single_site_magnon_energies(Jmix_qp, q_qc, muc) * 1.e3

# We separate the two magnon modes by sorting them
Esph1_qnr = np.sort(Esph1_qnr, axis=1)
Esph2_qnr = np.sort(Esph2_qnr, axis=1)

# First, we plot the magnon dispersion calculated with equivalent spherical
# sublattice site kernels
colors = rcParams['axes.prop_cycle'].by_key()['color']
# We define the lower bound on the magnon energy as the minimum within the
# chosen rc range, the "best estimate" as the median, and the upper bound
# as the maximum value
Evalid_qnr = Esph2_qnr[..., np.logical_and(rc_r >= rmin, rc_r <= rmax)]
Emin_qn = np.min(Evalid_qnr, axis=2)
E_qn = np.median(Evalid_qnr, axis=2)
Emax_qn = np.max(Evalid_qnr, axis=2)

# Plot one mode at a time
for n in range(2):
    plt.fill_between(pathq_q, Emin_qn[:, n], Emax_qn[:, n],
                     color=colors[0], alpha=0.4)
    plt.plot(pathq_q, Emin_qn[:, n], color='0.5')
    plt.plot(pathq_q, Emax_qn[:, n], color='0.5')
    if n == 0:
        label = 'eq. spheres'
    else:
        label = None
    plt.plot(pathq_q, E_qn[:, n], color=colors[0], label=label)

# Secondly, we plot the dispersion with inequivalent spherical sites
Eineq_qn = Esph1_qnr[..., np.where(np.abs(rc_r - rineq) < 1.e-8)[0][0]]
plt.plot(pathq_q, Eineq_qn[:, 0], '-o',
         mec='k', color=colors[1], label='ineq. spheres')
plt.plot(pathq_q, Eineq_qn[:, 1], '-o',
         mec='k', color=colors[1])

# Lastly plot the magnon dispersion with only a single site in the unit cell
plt.plot(pathq_q, Emix_qp[:, 0], '-v',
         mec='k', color=colors[2], label='unit cell')
plt.plot(pathq_q, Emix_qp[:, 1], '-^',
         mec='k', color=colors[3], label='cylinder')

# Use high-symmetry points as tickmarks for the x-axis
qticks = []
qticklabels = []
for sp, spq_c in zip(sp_p, spq_pc):
    for q, q_c in enumerate(q_qc):
        if np.allclose(q_c, spq_c):
            qticks.append(pathq_q[q])
            qticklabels.append(sp)
plt.xticks(qticks, qticklabels)
# Plot also vertical lines for each special point
for pq in qticks:
    plt.axvline(pq, color='0.5', linewidth=1, zorder=0)

# Labels and limits
plt.ylabel(mwlabel)
plt.xlim(qlim)
plt.ylim(mwlim)

plt.legend(title='Site geometry')

plt.savefig(filename, format=filename.split('.')[-1],
            bbox_inches='tight')



import os
import warnings
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pytest
from _pytest.tmpdir import _mk_tmp
from ase import Atoms
from ase.build import bulk
from ase.io import read
from gpaw import GPAW, PW, Davidson, FermiDirac, setup_paths
from gpaw.cli.info import info
from gpaw.mpi import broadcast, world
from gpaw.utilities import devnull


@contextmanager
def execute_in_tmp_path(request, tmp_path_factory):
    if world.rank == 0:
        # Obtain basename as
        # * request.function.__name__  for function fixture
        # * request.module.__name__    for module fixture
        basename = getattr(request, request.scope).__name__
        path = tmp_path_factory.mktemp(basename)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@pytest.fixture(scope='function')
def in_tmp_dir(request, tmp_path_factory):
    """Run test function in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='module')
def module_tmp_path(request, tmp_path_factory):
    """Run test module in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture
def add_cwd_to_setup_paths():
    """Temporarily add current working directory to setup_paths."""
    try:
        setup_paths[:0] = ['.']
        yield
    finally:
        del setup_paths[:1]


@pytest.fixture(scope='session')
def gpw_files(request, tmp_path_factory):
    """Reuse gpw-files.

    Returns a dict mapping names to paths to gpw-files.  If you
    want to reuse gpw-files from an earlier pytest session then set the
    ``$GPW_TEST_FILES`` environment variable and the files will be written
    to that folder.

    Example::

        def test_something(gpw_files):
            calc = GPAW(gpw_files['h2_lcao_wfs'])
            ...

    Possible systems are:

    * Bulk BCC-Li with 3x3x3 k-points: ``bcc_li_pw``, ``bcc_li_fd``,
      ``bcc_li_lcao``.

    * O2 molecule: ``o2_pw``.

    * H2 molecule: ``h2_pw``, ``h2_fd``, ``h2_lcao``.

    * H2 molecule (not centered): ``h2_pw_0``.

    * Spin-polarized H atom: ``h_pw``.

    * Polyethylene chain.  One unit, 3 k-points, no symmetry:
      ``c2h4_pw_nosym``.  Three units: ``c6h12_pw``.

    * Bulk TiO2 with 4x4x4 k-points: ``ti2o4_pw`` and ``ti2o4_pw_nosym``.

    * Bulk BN (zinkblende) with 2x2x2 k-points and 9 converged bands:
      ``bn_pw``.

    * Graphene with 6x6x1 k-points: graphene_pw

    * MoS2 with 6x6x1 k-points: mos2_pw

    Files with wave functions are also availabe (add ``_wfs`` to the names).
    """
    path = os.environ.get('GPW_TEST_FILES')
    if not path:
        warnings.warn(
            'Note that you can speed up the tests by reusing gpw-files '
            'from an earlier pytest session: '
            'set the $GPW_TEST_FILES environment variable and the '
            'files will be written to/read from that folder. '
            'See: https://wiki.fysik.dtu.dk/gpaw/devel/testing.html'
            '#gpaw.test.conftest.gpw_files')
        if world.rank == 0:
            path = _mk_tmp(request, tmp_path_factory)
        else:
            path = None
        path = broadcast(path)
    return GPWFiles(Path(path))


class GPWFiles:
    """Create gpw-files."""
    def __init__(self, path: Path):
        self.path = path
        path.mkdir(exist_ok=True)
        self.gpw_files = {}
        for file in path.glob('*.gpw'):
            self.gpw_files[file.name[:-4]] = file

    def __getitem__(self, name: str) -> Path:
        if name not in self.gpw_files:
            rawname, _, _ = name.partition('_wfs')
            calc = getattr(self, rawname)()
            path = self.path / (rawname + '.gpw')
            calc.write(path)
            self.gpw_files[rawname] = path
            path = self.path / (rawname + '_wfs.gpw')
            calc.write(path, mode='all')
            self.gpw_files[rawname + '_wfs'] = path
        return self.gpw_files[name]

    def bcc_li_pw(self):
        return self.bcc_li({'name': 'pw', 'ecut': 200})

    def bcc_li_fd(self):
        return self.bcc_li({'name': 'fd'})

    def bcc_li_lcao(self):
        return self.bcc_li({'name': 'lcao'})

    def bcc_li(self, mode):
        li = bulk('Li', 'bcc', 3.49)
        li.calc = GPAW(mode=mode,
                       kpts=(3, 3, 3),
                       txt=self.path / f'bcc_li_{mode["name"]}.txt')
        li.get_potential_energy()
        return li.calc

    def h2_pw(self):
        return self.h2({'name': 'pw', 'ecut': 200})

    def h2_fd(self):
        return self.h2({'name': 'fd'})

    def h2_lcao(self):
        return self.h2({'name': 'lcao'})

    def h2(self, mode):
        h2 = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]])
        h2.center(vacuum=2.5)
        h2.calc = GPAW(mode=mode,
                       txt=self.path / f'h2_{mode["name"]}.txt')
        h2.get_potential_energy()
        return h2.calc

    def h2_pw_0(self):
        h2 = Atoms('H2',
                   positions=[[-0.37, 0, 0], [0.37, 0, 0]],
                   cell=[5.74, 5, 5],
                   pbc=True)
        h2.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       txt=self.path / 'h2_pw_0.txt')
        h2.get_potential_energy()
        return h2.calc

    def h_pw(self):
        h = Atoms('H', magmoms=[1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 500},
                      txt=self.path / 'h_pw.txt')
        h.get_potential_energy()
        return h.calc

    def o2_pw(self):
        d = 1.1
        a = Atoms('O2', positions=[[0, 0, 0], [d, 0, 0]], magmoms=[1, 1])
        a.center(vacuum=4.0)
        a.calc = GPAW(mode={'name': 'pw', 'ecut': 800},
                      txt=self.path / 'o2_pw.txt')
        a.get_potential_energy()
        return a.calc

    def co_lcao(self):
        d = 1.1
        co = Atoms('CO', positions=[[0, 0, 0], [d, 0, 0]])
        co.center(vacuum=4.0)
        co.calc = GPAW(mode='lcao',
                       txt=self.path / 'co_lcao.txt')
        co.get_potential_energy()
        return co.calc

    def c2h4_pw_nosym(self):
        d = 1.54
        h = 1.1
        x = d * (2 / 3)**0.5
        z = d / 3**0.5
        pe = Atoms('C2H4',
                   positions=[[0, 0, 0],
                              [x, 0, z],
                              [0, -h * (2 / 3)**0.5, -h / 3**0.5],
                              [0, h * (2 / 3)**0.5, -h / 3**0.5],
                              [x, -h * (2 / 3)**0.5, z + h / 3**0.5],
                              [x, h * (2 / 3)**0.5, z + h / 3**0.5]],
                   cell=[2 * x, 0, 0],
                   pbc=(1, 0, 0))
        pe.center(vacuum=2.0, axis=(1, 2))
        pe.calc = GPAW(mode='pw',
                       kpts=(3, 1, 1),
                       symmetry='off',
                       txt=self.path / 'c2h4_pw_nosym.txt')
        pe.get_potential_energy()
        return pe.calc

    def c6h12_pw(self):
        pe = read(self['c2h4_pw_nosym'])
        pe = pe.repeat((3, 1, 1))
        pe.calc = GPAW(mode='pw', txt=self.path / 'c6h12_pw.txt')
        pe.get_potential_energy()
        return pe.calc

    def h2o_lcao(self):
        from ase.build import molecule
        atoms = molecule('H2O', cell=[8, 8, 8], pbc=1)
        atoms.center()
        atoms.calc = GPAW(mode='lcao', txt=self.path / 'h2o.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def ti2o4(self, symmetry):
        pwcutoff = 400.0
        k = 4
        a = 4.59
        c = 2.96
        u = 0.305

        rutile_cell = [[a, 0, 0],
                       [0, a, 0],
                       [0, 0, c]]

        TiO2_basis = np.array([[0.0, 0.0, 0.0],
                               [0.5, 0.5, 0.5],
                               [u, u, 0.0],
                               [-u, -u, 0.0],
                               [0.5 + u, 0.5 - u, 0.5],
                               [0.5 - u, 0.5 + u, 0.5]])

        bulk_crystal = Atoms(symbols='Ti2O4',
                             scaled_positions=TiO2_basis,
                             cell=rutile_cell,
                             pbc=(1, 1, 1))

        tag = '_nosym' if symmetry == 'off' else ''
        bulk_calc = GPAW(mode=PW(pwcutoff),
                         nbands=42,
                         eigensolver=Davidson(1),
                         kpts={'size': (k, k, k), 'gamma': True},
                         xc='PBE',
                         occupations=FermiDirac(0.00001),
                         parallel={'band': 1},
                         symmetry=symmetry,
                         txt=self.path / f'ti2o4_pw{tag}.txt')

        bulk_crystal.calc = bulk_calc
        bulk_crystal.get_potential_energy()
        return bulk_calc

    def ti2o4_pw(self):
        return self.ti2o4({})

    def ti2o4_pw_nosym(self):
        return self.ti2o4('off')

    def bn_pw(self):
        atoms = bulk('BN', 'zincblende', a=3.615)
        atoms.calc = GPAW(mode=PW(400),
                          kpts={'size': (2, 2, 2), 'gamma': True},
                          nbands=12,
                          convergence={'bands': 9},
                          occupations=FermiDirac(0.001),
                          txt=self.path / 'bn_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def graphene_pw(self):
        from ase.lattice.hexagonal import Graphene
        atoms = Graphene(symbol='C',
                         latticeconstant={'a': 2.45, 'c': 1.0},
                         size=(1, 1, 1))
        atoms.pbc = (1, 1, 0)
        atoms.center(axis=2, vacuum=4.0)
        ecut = 250
        nkpts = 6
        atoms.calc = GPAW(mode=PW(ecut),
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          nbands=len(atoms) * 6,
                          txt=self.path / 'graphene_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def mos2_pw(self):
        from ase.build import mx2
        atoms = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.127,
                    size=(1, 1, 1), vacuum=5)
        atoms.pbc = (1, 1, 1)
        ecut = 250
        nkpts = 6
        atoms.calc = GPAW(mode=PW(ecut),
                          xc='LDA',
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          occupations=FermiDirac(0.01),
                          txt=self.path / 'mos2_pw.txt')

        atoms.get_potential_energy()
        return atoms.calc


class GPAWPlugin:
    def __init__(self):
        if world.rank == -1:
            print()
            info()

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        from gpaw.mpi import size
        terminalreporter.section('GPAW-MPI stuff')
        terminalreporter.write(f'size: {size}\n')


def pytest_configure(config):
    if world.rank != 0:
        try:
            tw = config.get_terminal_writer()
        except AttributeError:
            pass
        else:
            tw._file = devnull
    config.pluginmanager.register(GPAWPlugin(), 'pytest_gpaw')
    for line in ['soc: Spin-orbit coupling',
                 'slow: slow test',
                 'fast: fast test',
                 'ci: test included in CI',
                 'libxc: LibXC requirered',
                 'mgga: MGGA test',
                 'dscf: Delta-SCF',
                 'mom: MOM',
                 'gllb: GLLBSC tests',
                 'elph: Electron-phonon',
                 'intel: fails on INTEL toolchain',
                 'response: tests of the response code',
                 'kspair: tests of kspair in the response code',
                 'serial: run in serial only',
                 'skip_for_new_gpaw: know failure for new refactored GPAW']:
        config.addinivalue_line('markers', line)


def pytest_runtest_setup(item):
    """Skip some tests.

    If:

    * they depend on libxc and GPAW is not compiled with libxc
    * they are before $PYTEST_START_AFTER
    """
    from gpaw import libraries

    if world.size > 1:
        for mark in item.iter_markers():
            if mark.name == 'serial':
                pytest.skip('Only run in serial')

    if item.location[0] <= os.environ.get('PYTEST_START_AFTER', ''):
        pytest.skip('Not after $PYTEST_START_AFTER')
        return

    if libraries['libxc']:
        return

    if any(mark.name in {'libxc', 'mgga'}
           for mark in item.iter_markers()):
        pytest.skip('No LibXC.')


@pytest.fixture
def scalapack():
    """Skip if not compiled with sl.

    This fixture otherwise does not return or do anything."""
    from gpaw.utilities import compiled_with_sl
    if not compiled_with_sl():
        pytest.skip(reason='no scalapack')



from gpaw import GPAW
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

# Calculate ground state with full unoccupied space
calc = GPAW('gs.gpw', txt=None).fixed_density(nbands='nao', txt='unocc.out')
calc.write('unocc.gpw', mode='all')

# Construct KS electron-hole basis
ksd = KohnShamDecomposition(calc)
ksd.initialize(calc)
ksd.write('ksd.ulm')



import numpy as np
from ase.dft.kpoints import monkhorst_pack


class GammaIntegrator:
    def __init__(self, truncation, kd, pd, chi0_wvv, chi0_wxvG):
        N = 4
        N_c = np.array([N, N, N])
        if truncation is not None:
            # Only average periodic directions if trunction is used
            N_c[kd.N_c == 1] = 1
        qf_qc = monkhorst_pack(N_c) / kd.N_c
        qf_qc *= 1.0e-6
        U_scc = kd.symmetry.op_scc
        qf_qc = kd.get_ibz_q_points(qf_qc, U_scc)[0]
        self.weight_q = kd.q_weights
        self.qf_qv = 2 * np.pi * (qf_qc @ pd.gd.icell_cv)
        self.a_wq = np.sum([chi0_vq * self.qf_qv.T
                            for chi0_vq in
                            np.dot(chi0_wvv, self.qf_qv.T)],
                           axis=1)
        self.a0_qwG = np.dot(self.qf_qv, chi0_wxvG[:, 0])
        self.a1_qwG = np.dot(self.qf_qv, chi0_wxvG[:, 1])



scalapack = True

extra_compile_args = ['-O3', '-std=c99', '-fpic']

compiler = 'gcc'
mpicompiler = '/opt/apps/gcc4_4/openmpi/1.3b/bin/mpicc'
mpilinker = mpicompiler

mkl_lib_path = '/opt/apps/intel/mkl/10.0.1.014/lib/em64t/'
ompi_lib_path = '/opt/apps/gcc4_4/openmpi/1.3b/lib'

libraries = []

# use static linking to avoid
# "cannot allocate memory for thread-local data: ABORT"
extra_link_args = [
mkl_lib_path+'libmkl_intel_lp64.a',
mkl_lib_path+'libmkl_sequential.a',
mkl_lib_path+'libmkl_core.a',
mkl_lib_path+'libmkl_blacs_openmpi_lp64.a',
mkl_lib_path+'libmkl_scalapack_lp64.a',
mkl_lib_path+'libmkl_blacs_openmpi_lp64.a',
mkl_lib_path+'libmkl_intel_lp64.a',
mkl_lib_path+'libmkl_sequential.a',
mkl_lib_path+'libmkl_core.a',
mkl_lib_path+'libmkl_intel_lp64.a',
mkl_lib_path+'libmkl_sequential.a',
mkl_lib_path+'libmkl_core.a',
]

extra_link_args += ['-Wl,-rpath='+ompi_lib_path]

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



#!/usr/bin/env python
"""Transport exersice

This file should do the same as pt_h2_lcao.py, but extracts the Hamiltonians
manually instead of using gpawtransport, which currently does not work
"""

from ase import Atoms
from gpaw import GPAW, Mixer, FermiDirac
from gpaw.lcao.tools import (remove_pbc,
                             get_lcao_hamiltonian,
                             get_lead_lcao_hamiltonian)
import pickle as pickle

a = 2.41  # Pt binding length
b = 0.90  # H2 binding length
c = 1.70  # Pt-H binding length
L = 7.00  # width of unit cell

#####################
# Scattering region #
#####################

# Setup the Atoms for the scattering region.
atoms = Atoms('Pt5H2Pt5', pbc=(1, 0, 0), cell=[9 * a + b + 2 * c, L, L])
atoms.positions[:5, 0] = [i * a for i in range(5)]
atoms.positions[-5:, 0] = [i * a + b + 2 * c for i in range(4, 9)]
atoms.positions[5:7, 0] = [4 * a + c, 4 * a + c + b]
atoms.positions[:, 1:] = L / 2.

# Attach a GPAW calculator
calc = GPAW(h=0.3,
            xc='PBE',
            basis='szp(dzp)',
            occupations=FermiDirac(width=0.1),
            kpts=(1, 1, 1),
            mode='lcao',
            txt='pt_h2_lcao_scat.txt',
            mixer=Mixer(0.1, 5, weight=100.0),
            symmetry={'point_group': False, 'time_reversal': False})
atoms.calc = calc

atoms.get_potential_energy()  # Converge everything!
Ef = atoms.calc.get_fermi_level()

H_skMM, S_kMM = get_lcao_hamiltonian(calc)
# Only use first kpt, spin, as there are no more
H, S = H_skMM[0, 0], S_kMM[0]
H -= Ef * S
remove_pbc(atoms, H, S, 0)

# Dump the Hamiltonian and Scattering matrix to a pickle file
pickle.dump((H.astype(complex), S.astype(complex)),
            open('scat_hs.pickle', 'wb'), 2)

########################
# Left principal layer #
########################

# Use four Pt atoms in the lead, so only take those from before
atoms = atoms[:4].copy()
atoms.set_cell([4 * a, L, L])

# Attach a GPAW calculator
calc = GPAW(h=0.3,
            xc='PBE',
            basis='szp(dzp)',
            occupations=FermiDirac(width=0.1),
            kpts=(4, 1, 1),  # More kpts needed as the x-direction is shorter
            mode='lcao',
            txt='pt_h2_lcao_llead.txt',
            mixer=Mixer(0.1, 5, weight=100.0),
            symmetry={'point_group': False, 'time_reversal': False})
atoms.calc = calc

atoms.get_potential_energy()  # Converge everything!
Ef = atoms.calc.get_fermi_level()

ibz2d_k, weight2d_k, H_skMM, S_kMM = get_lead_lcao_hamiltonian(calc)
# Only use first kpt, spin, as there are no more
H, S = H_skMM[0, 0], S_kMM[0]
H -= Ef * S

# Dump the Hamiltonian and Scattering matrix to a pickle file
pickle.dump((H, S), open('lead1_hs.pickle', 'wb'), 2)

#########################
# Right principal layer #
#########################
# This is identical to the left prinicpal layer so we don't have to do anything
# Just dump the same Hamiltonian and Scattering matrix to a pickle file
pickle.dump((H, S), open('lead2_hs.pickle', 'wb'), 2)



# %%
"""
# Discussion
"""

# %%
"""
Now, we will finally visualize the absorption spectra we have calculated. The script below is setup for plotting and saving the absorption spectra for BN with 24x24x1 k-points. First we load the x, y, and z components and plot all three curves in different colors.
"""

# %%
#Plot the absorption spectrum:
import numpy as np
import matplotlib.pyplot as plt
plt.figure()

#Of course you need to input the name you gave the files earlier own yourself and only plot the components you calculated.
#Here is only showed hot to plot the x-component.

absox = np.loadtxt('CdTe_rpa_x.csv', delimiter=',') # student: absox = np.loadtxt('???_rpa_x.csv', delimiter=',')
plt.plot(absox[:, 0], absox[:, 4], label='RPA_x CdTe 12x12x4', lw=2, color='b')

plt.xlabel(r'$\hbar\omega\;[eV]$', size=20)
plt.ylabel(r'$Im(\epsilon)$', size=20)
plt.xticks(size=16)
plt.yticks(size=16)
plt.tight_layout()
plt.axis([0.0, 10.0, None, None])
plt.legend()

plt.show() 


# %%
"""
Now open the saved figure and inspect the absorption spectra. Does it look as you expected? Can you guess the band gap from the absorption spectra and which of the band gaps calculated on day 2 does it match? 

Now plot the absorption spectra for the other k-point meshes you calculated and compare. You could for instance modify the script above to plot, in different colors, all spectra in the same plot for comparison. Would you say that the calculation is converged? 

Finally, talk with the other groups and compare your absorption spectra with their absorption spectra. Is the same number of k-points needed to obtain same degree of convergence for the different materials? Is there any difference for the 2D material (Boron-Nitride)?

For a more realistic description one would have to include excitonic effects (i.e. the electron-hole interaction). As an outlook you might consider reading:
https://wiki.fysik.dtu.dk/gpaw/documentation/bse/bse.html
"""


from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

a = bulk('Pt', 'fcc', a=3.92)

calc = GPAW(mode=PW(600),
            xc='PBE',
            nbands=20,
            parallel={'domain': 1, 'band': 1},
            occupations=FermiDirac(width=0.01),
            convergence={'density': 1.e-6},
            kpts=[8, 8, 8],
            txt='Pt_gs.txt')

a.calc = calc
a.get_potential_energy()

calc.write('Pt_gs.gpw', mode='all')



import pytest
from gpaw.core import UniformGrid
from gpaw.mpi import world


@pytest.mark.ci
def test_redist():
    a = 2.5
    n = 2

    # comm = world.new_communicator([world.rank])
    grid1 = UniformGrid(cell=[a, a, a], size=(n, n, n), comm=world)
    f1 = grid1.empty()
    f1.data[:] = world.rank + 1
    f2 = f1.gather()
    f3 = f1.gather(broadcast=True)
    if world.rank == 0:
        assert (f2.data == f3.data).all()
    print(f2)
    f4 = f1.new()
    f4.scatter_from(f2)
    assert (f4.data == f1.data).all()



# creates: theme.css
import re
import sphinx_rtd_theme


rtd_path = sphinx_rtd_theme.get_html_theme_path()
with open(rtd_path + '/sphinx_rtd_theme/static/css/theme.css') as fd:
    css = fd.read()


def replace(match):
    s = match.group()
    if len(s) == 8:
        return s[0:3] + s[5:7] + s[3:5] + s[7]
    if len(s) == 5:
        return s[0:2] + s[3] + s[2] + s[4]
    1 / 0


new = re.sub('#[a-fA-F0-9]{3,6}[;}]', replace, css)
with open('theme.css', 'w') as fd:
    fd.write(new)
    # Equation numbers to the right:
    fd.write('span.eqno {float: right;}\n')



from ase.build import molecule
from ase.optimize import QuasiNewton
from gpaw import GPAW

H2 = molecule('H2', cell=(10, 10, 10))
H2.center()
calc = GPAW()
H2.calc = calc
dyn = QuasiNewton(H2, trajectory='H2.traj')
dyn.run(fmax=0.05)



# web-page: field.ind_Ffe.png
from gpaw.mpi import world

import numpy as np
import matplotlib.pyplot as plt

from gpaw.inducedfield.inducedfield_base import BaseInducedField
from gpaw.tddft.units import aufrequency_to_eV

assert world.size == 1, 'This script should be run in serial mode.'


# Helper function
def do_plot(d_g, ng, box, atoms):
    # Take slice of data array
    d_yx = d_g[:, :, ng[2] // 2]
    y = np.linspace(0, box[0], ng[0] + 1)[:-1]
    dy = box[0] / (ng[0] + 1)
    y += dy * 0.5
    ylabel = u'x / Å'
    x = np.linspace(0, box[1], ng[1] + 1)[:-1]
    dx = box[1] / (ng[1] + 1)
    x += dx * 0.5
    xlabel = u'y / Å'

    # Plot
    plt.figure()
    ax = plt.subplot(1, 1, 1)
    X, Y = np.meshgrid(x, y)
    plt.contourf(X, Y, d_yx, 40)
    plt.colorbar()
    for atom in atoms:
        pos = atom.position
        plt.scatter(pos[1], pos[0], s=50, c='k', marker='o')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xlim([x[0], x[-1]])
    plt.ylim([y[0], y[-1]])
    ax.set_aspect('equal')


for fname, name in zip(['field.ind'], ['Classical system']):
    # Read InducedField object
    ind = BaseInducedField(fname, readmode='all')

    # Choose array
    w = 0                                      # Frequency index
    freq = ind.omega_w[w] * aufrequency_to_eV  # Frequency
    box = np.diag(ind.atoms.get_cell())        # Calculation box
    d_g = ind.Ffe_wg[w]                        # Data array
    ng = d_g.shape                             # Size of grid
    atoms = ind.atoms                          # Atoms

    do_plot(d_g, ng, box, atoms)
    plt.title(f'{name}\nField enhancement @ {freq:.2f} eV')
    plt.savefig(fname + '_Ffe.png', bbox_inches='tight')

    # Imaginary part of density
    d_g = ind.Frho_wg[w].imag
    ng = d_g.shape
    do_plot(d_g, ng, box, atoms)
    plt.title('%s\nImaginary part of induced charge density @ %.2f eV' %
              (name, freq))
    plt.savefig(fname + '_Frho.png', bbox_inches='tight')

    # Imaginary part of potential
    d_g = ind.Fphi_wg[w].imag
    ng = d_g.shape
    do_plot(d_g, ng, box, atoms)
    plt.title(f'{name}\nImaginary part of induced potential @ {freq:.2f} eV')
    plt.savefig(fname + '_Fphi.png', bbox_inches='tight')



extra_compile_args = ['-std=c99', '-O3'] #, '-O0']
compiler = 'cc'
mpicompiler = 'cc'
mpilinker= 'cc'
libraries = ['acml']
extra_link_args += ['-dynamic']
include_dirs += ['/usr/lib64/python2.6/site-packages/numpy/core/include']
# include_dirs += ['/path_in_workspace/lib/python/numpy/core/include/']

scalapack = True
hdf5 = True

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
define_macros += [("GPAW_ASYNC",1)]
define_macros += [("GPAW_MPI2",1)]
define_macros += [("GPAW_PERFORMANCE_REPORT",1)]



# web-page: N2Ru_hollow.png, 2NadsRu.png, TS.xyz
from myqueue.workflow import run


def workflow():
    with run(script='check_convergence.py', tmax='5h', cores=8):
        run(script='convergence.py')

    with run(script='n2_on_metal.py', tmax='6h'):
        with run(script='neb.py', tmax='3h', cores=8):
            run(script='vibrations.py', tmax='9h', cores=8)



"""Test self-consistency of the extend_head flag for a semi-conductor."""

# General modules
import pytest
import numpy as np
from pathlib import Path

# Script modules
from ase import Atoms
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack

from gpaw import GPAW, PW
from gpaw.mpi import world
from gpaw.response.chi0 import Chi0


# ---------- Chi0 parametrization ---------- #


def generate_semic_chi0_params():
    """Check the following options for a semi-conductor:
    * threshold
    * hilbert
    * timeordered
    * nbands
    * eta=0.
    * real_space_derivatives
    """
    # Define default parameters
    chi0kwargs = dict(
        frequencies=np.linspace(0., 30., 11),
        eta=0.05,
        hilbert=False,
        timeordered=False,
        threshold=1,
        real_space_derivatives=False)
    chi0_params = [chi0kwargs]

    # Check different chi0 parameter combinations
    ck1 = chi0kwargs.copy()  # Check k.p threshold
    ck1['threshold'] = 0.5
    chi0_params.append(ck1)

    ck2 = chi0kwargs.copy()  # Check hilbert transform
    ck2['hilbert'] = True
    ck2['frequencies'] = None
    chi0_params.append(ck2)

    ck3 = chi0kwargs.copy()  # Check timeordering
    ck3['timeordered'] = True
    chi0_params.append(ck3)

    ck4 = chi0kwargs.copy()  # Check nbands
    ck4['nbands'] = None
    chi0_params.append(ck4)

    ck5 = chi0kwargs.copy()  # Check eta=0.
    ck5['frequencies'] = 1.j * ck5['frequencies'][1:]
    ck5['eta'] = 0.
    chi0_params.append(ck5)

    ck6 = chi0kwargs.copy()  # Check real space derivs.
    ck6['real_space_derivatives'] = True
    chi0_params.append(ck6)

    return chi0_params


def generate_metal_chi0_params():
    """In addition to the semi-conductor parameters, test also:
    * integrationmode
    * intraband (test all other settings with and without intraband)
    """
    # Get semi-conductor defaults
    chi0_params = generate_semic_chi0_params()
    chi0kwargs = chi0_params[0]

    ck7 = chi0kwargs.copy()  # Check tetrahedron integration
    ck7['integrationmode'] = 'tetrahedron integration'
    ck7['hilbert'] = True
    ck7['frequencies'] = None
    chi0_params.append(ck7)

    # Run all test settings without intraband
    nointra_chi0_params = []
    for ck in chi0_params:
        ck['intraband'] = True  # This is the default, but be specific
        ck['rate'] = 0.1  # We need a finite scattering rate, not to blow up
        nointra_ck = ck.copy()
        nointra_ck['intraband'] = False
        nointra_ck['rate'] = 0.0
        nointra_chi0_params.append(nointra_ck)
    chi0_params += nointra_chi0_params

    return chi0_params


def generate_magnetic_metal_chi0_params():
    """In addition to the magnetic parametrization, test also:
    * spin
    """
    # Get metallic defaults
    chi0_params = generate_metal_chi0_params()
    chi0kwargs = chi0_params[0]

    cks0 = chi0kwargs.copy()  # Check spin up
    cks0['spins'] = [0]
    cks0['intraband'] = False
    chi0_params.append(cks0)

    cks1 = chi0kwargs.copy()  # Check spin down
    cks1['spins'] = [1]
    cks1['intraband'] = False
    chi0_params.append(cks1)

    return chi0_params


@pytest.fixture(scope='module', params=generate_semic_chi0_params())
def He_chi0kwargs(request, He_gs):
    chi0kwargs = request.param
    assure_nbands(chi0kwargs, He_gs)

    return chi0kwargs


@pytest.fixture(scope='module', params=generate_metal_chi0_params())
def Li_chi0kwargs(request, Li_gs):
    chi0kwargs = request.param
    assure_nbands(chi0kwargs, Li_gs)

    return chi0kwargs


@pytest.fixture(scope='module', params=generate_magnetic_metal_chi0_params())
def Ni_chi0kwargs(request, Ni_gs):
    chi0kwargs = request.param
    assure_nbands(chi0kwargs, Ni_gs)

    return chi0kwargs


def assure_nbands(chi0kwargs, my_gs):
    # Fill in nbands parameter, if not already specified
    if 'nbands' not in chi0kwargs:
        _, nbands = my_gs
        chi0kwargs['nbands'] = nbands


# ---------- Actual tests ---------- #


@pytest.mark.response
def test_he_chi0_extend_head(in_tmp_dir, He_gs, He_chi0kwargs):
    chi0_extend_head_test(He_gs, He_chi0kwargs)


@pytest.mark.response
def test_li_chi0_extend_head(in_tmp_dir, Li_gs, Li_chi0kwargs, request):
    mark_tetrahedron_xfail(Li_chi0kwargs, request)
    chi0_extend_head_test(Li_gs, Li_chi0kwargs)


@pytest.mark.response
def test_ni_chi0_extend_head(in_tmp_dir, Ni_gs, Ni_chi0kwargs, request):
    mark_tetrahedron_xfail(Ni_chi0kwargs, request)
    chi0_extend_head_test(Ni_gs, Ni_chi0kwargs)


def mark_tetrahedron_xfail(chi0kwargs, request):
    if ('integrationmode' in chi0kwargs and
        chi0kwargs['integrationmode'] == 'tetrahedron integration'):
        # Head and wings have not yet have a tetrahedron integration
        # implementation. For now, we simply mark the tests with xfail.
        request.node.add_marker(pytest.mark.xfail)


def chi0_extend_head_test(my_gs, chi0kwargs):
    # ---------- Inputs ---------- #
    gpw, nbands = my_gs

    ecut = 50

    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # ---------- Script ---------- #

    if 'spins' in chi0kwargs:
        spins = chi0kwargs.pop('spins')
    else:
        spins = 'all'

    chi0fac = Chi0(gpw, ecut=ecut,
                   nblocks=nblocks,
                   **chi0kwargs)

    chi0_f = calculate_optical_limit(chi0fac, spins=spins, extend_head=False)
    chi0_t = calculate_optical_limit(chi0fac, spins=spins, extend_head=True)

    is_equal, err_msg = chi0_data_is_equal(chi0_f, chi0_t)
    assert is_equal, err_msg


# ---------- System ground states ---------- #


@pytest.fixture(scope='module')
def He_gs(module_tmp_path):
    # ---------- Inputs ---------- #

    a = 3.0
    xc = 'LDA'
    kpts = 2
    nbands = 1 + 1 + 3  # 1s + 2s + 2p empty shell bands
    ebands = 1  # Include also 3s bands for numerical consistency
    pw = 250
    convergence = {'bands': nbands}
    gpw = Path('He.gpw').resolve()

    # ---------- Script ---------- #

    atoms = Atoms('He', cell=[a, a, a], pbc=True)
    calculate_gs(atoms, gpw, pw, kpts, nbands, ebands,
                 xc=xc, convergence=convergence)

    return gpw, nbands


@pytest.fixture(scope='module')
def Li_gs(module_tmp_path):
    # ---------- Inputs ---------- #

    a = 3.49
    xc = 'LDA'
    kpts = 3
    nbands = 1 + 3 + 1  # 2s + 2p + 3s empty shell bands
    ebands = 3  # Include also 3p bands for numerical consistency
    pw = 250
    convergence = {'bands': nbands}
    gpw = Path('Li.gpw').resolve()

    # ---------- Script ---------- #

    atoms = bulk('Li', 'bcc', a=a)
    calculate_gs(atoms, gpw, pw, kpts, nbands, ebands,
                 xc=xc, convergence=convergence)

    return gpw, nbands


@pytest.fixture(scope='module')
def Ni_gs(module_tmp_path):
    # ---------- Inputs ---------- #

    a = 3.524
    xc = 'LDA'
    kpts = 4
    setups = {'Ni': '10'}
    nbands = 1 + 5  # 4s + 3d valence bands
    ebands = 3 + 1  # Include also 4p and 5s for numerical consistency
    pw = 250
    convergence = {'bands': nbands}
    gpw = Path('Ni.gpw').resolve()

    # ---------- Script ---------- #

    atoms = bulk('Ni', 'fcc', a=a)
    atoms.set_initial_magnetic_moments([0.6])
    calculate_gs(atoms, gpw, pw, kpts, nbands, ebands,
                 setups=setups, xc=xc, convergence=convergence)

    return gpw, nbands


# ---------- Script functionality ---------- #


def calculate_gs(atoms, gpw, pw, kpts, nbands, ebands,
                 **kwargs):
    calc = GPAW(mode=PW(pw),
                kpts=monkhorst_pack((kpts, kpts, kpts)),
                nbands=nbands + ebands,
                symmetry={'point_group': True},
                parallel={'domain': 1},
                **kwargs)

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write(gpw, 'all')


def calculate_optical_limit(chi0_factory, spins='all', extend_head=True):
    """Use the update_chi0 method to calculate chi0 for q=0."""
    # Create Chi0Data object
    chi0 = chi0_factory.create_chi0([0., 0., 0.], extend_head=extend_head)

    # Prepare update_chi0
    if spins == 'all':
        spins = range(chi0_factory.calc.wfs.nspins)
    m1 = chi0_factory.nocc1
    m2 = chi0_factory.nbands
    chi0_factory.plasmafreq_vv = np.zeros((3, 3), complex)

    chi0 = chi0_factory.update_chi0(chi0, m1, m2, spins)

    # Distribute over frequencies
    chi0.chi0_wGG = chi0.distribute_frequencies()

    return chi0


def chi0_data_is_equal(chi0_f, chi0_t):
    """Compare Chi0Data objects without and with extend_head=True."""
    err_msg = None
    if not np.allclose(chi0_f.chi0_wGG, chi0_t.chi0_wGG):
        err_msg = 'body mismatch'
    elif not np.allclose(chi0_f.chi0_wxvG, chi0_t.chi0_wxvG):
        err_msg = 'wings mismatch'
    elif not np.allclose(chi0_f.chi0_wvv, chi0_t.chi0_wvv):
        err_msg = 'head mismatch'

    is_equal = err_msg is None

    return is_equal, err_msg



from ase import optimize
from ase.build import molecule
from ase.vibrations.infrared import Infrared
from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster

h = 0.20

atoms = Cluster(molecule('H2O'))
atoms.minimal_box(4, h=h)

# relax the molecule
calc = GPAW(xc='PBE', h=h, occupations=FermiDirac(width=0.1))
atoms.calc = calc

dyn = optimize.FIRE(atoms)
dyn.run(fmax=0.01)
atoms.write('relaxed.traj')

# finite displacement for vibrations
atoms.calc.set(symmetry={'point_group': False})
ir = Infrared(atoms, name='ir')
ir.run()



"""Tool for generating graps of objects."""
from pathlib import Path
from typing import Any

import ase
import numpy as np
from gpaw.core.atom_arrays import AtomArraysLayout
from gpaw.new.ase_interface import GPAW
from gpaw.new.brillouin import BZPoints
from gpaw.new.builder import builder


def create_nodes(obj, *objects, include):
    node1 = create_node(obj, include)
    nodes = {node.name: node for node in node1.nodes()}
    for obj in objects:
        for node in create_node(obj, include).nodes():
            if node.name not in nodes:
                nodes[node.name] = node

    for name, node in nodes.items():
        node.subclasses = []
        for cls in node.obj.__class__.__subclasses__():
            if cls.__name__ in nodes:
                node.subclasses.append(nodes[cls.__name__])

    for _ in range(1):
        new = {}
        for name, node in nodes.items():
            bases = node.obj.__class__.__bases__
            (cls,) = bases
            if cls is not object:
                base = nodes.get(cls.__name__)
                base = base or new.get(cls.__name__)
                if base is None:
                    base = Node(node.obj, node.attrs, [], None)
                    base.name = cls.__name__
                    base.has = node.has.copy()
                    new[cls.__name__] = base
                if node not in base.subclasses:
                    base.subclasses.append(node)
                if node is not base:
                    node.base = base

        for name, node in new.items():
            nodes[name] = node

    for node in nodes.values():
        node.fix()

    return list(nodes.values())


def create_node(obj, include):
    attrs = []
    arrows = []
    for key, value in obj.__dict__.items():
        if key[0] != '_':
            if not include(value):
                attrs.append(key)
            else:
                arrows.append(key)
    return Node(obj, attrs, arrows, include)


class Node:
    def __init__(self, obj, attrs, arrows, include):
        self.obj = obj
        self.name = obj.__class__.__name__
        self.attrs = attrs
        self.has = {key: create_node(getattr(obj, key), include)
                    for key in arrows}
        self.base = None
        self.subclasses = []
        self.rgb = None

    def __repr__(self):
        return (f'Node({self.name}, {self.attrs}, {list(self.has)}, ' +
                f'{self.base.name if self.base is not None else None}, ' +
                f'{[o.name for o in self.subclasses]})')

    def nodes(self):
        yield self
        for node in self.has.values():
            yield from node.nodes()

    def keys(self):
        return set(self.attrs + list(self.has))

    def superclass(self):
        return self if self.base is None else self.base.superclass()

    def fix(self):
        if self.subclasses:
            if len(self.subclasses) > 1:
                keys = self.subclasses[0].keys()
                for node in self.subclasses[1:]:
                    keys &= node.keys()
            else:
                keys = self.keys()
            self.attrs = [attr for attr in self.attrs if attr in keys]
            self.has = {key: value for key, value in self.has.items()
                        if key in keys}
            for obj in self.subclasses:
                obj.attrs = [attr for attr in obj.attrs if attr not in keys]
                obj.has = {key: value for key, value in obj.has.items()
                           if key not in keys}

    def color(self, rgb):
        self.rgb = rgb
        for obj in self.subclasses:
            obj.color(rgb)

    def plot(self, g):
        kwargs = {'style': 'filled',
                  'fillcolor': self.rgb} if self.rgb else {}
        if self.attrs:
            a = r'\n'.join(add_type(attr, getattr(self.obj, attr))
                           for attr in self.attrs)
            txt = f'{{{self.name} | {a}}}'
        else:
            txt = self.name
        g.node(self.name, txt, **kwargs)


def add_type(name: str, obj: Any) -> str:
    type = obj.__class__.__name__
    return f'{name}: {type}'


def plot_graph(figname, nodes, colors={}, replace={}):
    import graphviz
    g = graphviz.Digraph(node_attr={'shape': 'record'})

    for node in nodes:
        if node.name in colors:
            node.color(colors[node.name])

    for node in nodes:
        node.plot(g)
        for key, value in node.has.items():
            key = replace.get(key, key)
            g.edge(node.name, value.superclass().name, label=key)
        if node.base:
            g.edge(node.base.name, node.name, arrowhead='onormal')

    g.render(figname, format='svg')

    try:
        Path(figname).unlink()  # remove "dot" file
    except FileNotFoundError:
        pass


def abc():
    class A:
        def __init__(self, b):
            self.a = 1
            self.b = b

        def m(self):
            pass

    class B:
        pass

    class C(B):
        pass

    nodes = create_nodes(
        A(C()), B(),
        include=lambda obj: obj.__class__.__name__ in 'ABC')
    plot_graph('abc', nodes, {'B': '#ffddff'})


def code():
    fd = GPAW(mode='fd', txt=None)
    pw = GPAW(mode='pw', txt=None)
    lcao = GPAW(mode='lcao', txt=None)
    a = ase.Atoms('H', cell=[2, 2, 2], pbc=1)

    class Atoms:
        def __init__(self, calc):
            self.calc = calc

    a0 = Atoms(fd)
    fd.get_potential_energy(a)
    pw.get_potential_energy(a)
    lcao.get_potential_energy(a)
    ibzwfs = fd.calculation.state.ibzwfs
    ibzwfs.wfs_qs = ibzwfs.wfs_qs[0][0]

    colors = {'BZPoints': '#ddffdd',
              'PotentialCalculator': '#ffdddd',
              'WaveFunctions': '#ddddff',
              'Eigensolver': '#ffffdd',
              'PoissonSolver': '#ffeedd',
              'Hamiltonian': '#eeeeee'}

    def include(obj):
        try:
            mod = obj.__module__
        except AttributeError:
            return False

        return mod.startswith('gpaw.new')

    things = [pw, lcao,
              lcao.calculation.state.ibzwfs.wfs_qs[0][0],
              BZPoints(np.zeros((5, 3)))]
    nodes = create_nodes(a0, *things, include=include)
    plot_graph('code', nodes, colors,
               replace={'wfs_qs': 'wfs_qs[q][s]'})

    # scf.svg:
    nodes = create_nodes(
        fd.calculation.pot_calc.nct_aR,
        pw.calculation.pot_calc.nct_ag,
        include=lambda obj: obj.__class__.__name__.startswith('Atom'))
    plot_graph('acf', nodes, {'AtomCenteredFunctions': '#ddffff'})

    # da.svg:
    nodes = create_nodes(
        fd.calculation.state.ibzwfs.wfs_qs.psit_nX,
        pw.calculation.state.ibzwfs.wfs_qs[0][0].psit_nX,
        include=lambda obj:
            getattr(obj, '__module__', '').startswith('gpaw.core') and
            obj.__class__.__name__ != '_lru_cache_wrapper')
    plot_graph('da', nodes, {'DistributedArrays': '#eeeeee',
                             'Domain': '#dddddd'})


def builders():
    b = []
    a = ase.Atoms('H', cell=[2, 2, 2], pbc=1)
    for mode in ['fd', 'pw', 'lcao']:
        b.append(builder(a, {'mode': mode}))
    nodes = create_nodes(
        *b,
        include=lambda obj:
            obj.__class__.__name__.endswith('Builder') or
            obj.__class__.__name__ == 'InputParameters')
    plot_graph('builder', nodes, {'DFTComponentsBuilder': '#ffeedd'})


def aa():
    nodes = create_nodes(
        AtomArraysLayout([1]).empty(),
        include=lambda obj: obj.__class__.__name__.startswith('Atom'))
    nodes = [node for node in nodes if node.name != 'AtomArrays']
    for node in nodes:
        print(node)
        print(node.has)
        if node.name == 'DistributedArrays':
            node.name = 'AtomArrays'
            node.attrs.remove('dv')
            break
    plot_graph('aa', [node for node in nodes if node.name[0] == 'A'])


def main():
    abc()
    code()
    builders()
    aa()


if __name__ == '__main__':
    main()



from myqueue.workflow import run


def workflow():
    runs = [run(script='run.py', args=[cores], cores=cores, tmax=tmax)
            for cores, tmax in [(1, '1h'),
                                (4, '5h'),
                                (8, '12h'),
                                (16, '10h')]]
    run(script='analyse.py', deps=runs)



# creates: intro/intro.ipynb
# creates: batteries/batteries1.ipynb
# creates: batteries/batteries2.ipynb
# creates: batteries/batteries3.ipynb
# creates: catalysis/n2_on_metal.ipynb, catalysis/neb.ipynb
# creates: catalysis/vibrations.ipynb, catalysis/convergence.ipynb
# creates: magnetism/magnetism1.ipynb, magnetism/magnetism2.ipynb
# creates: magnetism/magnetism3.ipynb,
# creates: machinelearning/machinelearning.ipynb
# creates: photovoltaics/pv1.ipynb, photovoltaics/pv2.ipynb
# creates: photovoltaics/pv3.ipynb
from pathlib import Path
from gpaw.utilities.nbrun import py2ipynb
import sys

teacher = False
if len(sys.argv) == 2:
    if sys.argv[1] == '--teacher':
        teacher = True
    else:
        raise ValueError(f'Unknown command line argument: {sys.argv[1]}')

kernel = {'display_name': 'CAMD2022',
          'language': 'python',
          'name': 'camd2022'}

for path in Path().glob('*/*.py'):
    if path.read_text().startswith('# %%\n'):
        print(path)
        py2ipynb(path, kernel=kernel, teachermode=teacher)



import pytest
from ase import Atoms
from gpaw.new.ase_interface import GPAW

a = 2.5
k = 4


def test_afm_h_chain(in_tmp_dir):
    """Compare 2*H AFM cell with 1*H q=1/2 spin-spiral cell."""
    h = Atoms('H',
              magmoms=[1],
              cell=[a, 0, 0],
              pbc=[1, 0, 0])
    h.center(vacuum=2.0, axis=(1, 2))
    h.calc = GPAW(mode={'name': 'pw',
                        'ecut': 400,
                        'qspiral': [0.5, 0, 0]},
                  magmoms=[[1, 0, 0]],
                  symmetry='off',
                  kpts=(2 * k, 1, 1))
    e1 = h.get_potential_energy()
    h1, l1 = h.calc.get_homo_lumo()
    h.calc.write('h.gpw')
    print(e1, h.get_magnetic_moment())
    a1 = GPAW('h.gpw').get_atoms()
    print(a1.get_potential_energy(), a1.calc.calculation.magmoms())

    h2 = Atoms('H2',
               [(0, 0, 0), (a, 0, 0)],
               magmoms=[1, -1],
               cell=[2 * a, 0, 0],
               pbc=[1, 0, 0])
    h2.center(vacuum=2.0, axis=(1, 2))
    h2.calc = GPAW(mode={'name': 'pw',
                         'ecut': 400},
                   kpts=(k, 1, 1))
    e2 = h2.get_potential_energy()
    h2, l2 = h2.calc.get_homo_lumo()

    assert 2 * e1 == pytest.approx(e2, abs=0.002)
    assert h1 == pytest.approx(h2, abs=0.001)
    assert l1 == pytest.approx(l2, abs=0.001)



"""Ensure that all effective volumes are the same in graphene"""
import numpy as np
from ase import Atoms

from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw import GPAW, FermiDirac


c = 3
a = 1.42
atoms = Atoms(
    'C2', [(0, 0, 0.5), (1 / 3, 1 / 3, 0.5)], pbc=(1, 1, 0))
atoms.set_cell(
    [(np.sqrt(3) * a / 2.0, 3 / 2.0 * a, 0),
     (-np.sqrt(3) * a / 2.0, 3 / 2.0 * a, 0),
     (0, 0, 2 * c)],
    scale_atoms=True)
graphene = atoms.repeat([1, 2, 1])

h = 0.25
calc = GPAW(
    h=h,
    occupations=FermiDirac(0.1))
graphene.calc = calc
graphene.get_potential_energy()
vol_a = HirshfeldPartitioning(calc).get_effective_volume_ratios()
print(vol_a)
assert vol_a.ptp() < 1e-2



from pylab import gcf, gca, draw_if_interactive


def twiny(ay=None):
    """
    Make a second axes overlay ay (or the current axes if ay is None)
    sharing the yaxis.  The ticks for ay2 will be placed on the top,
    and the ay2 instance is returned.  See examples/two_scales.py
    """
    if ay is None:
        ay = gca()

    ay2 = gcf().add_axes(ay.get_position(), sharey=ay, frameon=False)
    ay2.xaxis.tick_top()
    ay2.xaxis.set_label_position('top')
    ay.xaxis.tick_bottom()
    draw_if_interactive()
    return ay2



import numpy as np
import pytest
from gpaw.core.matrix import Matrix
from gpaw.mpi import broadcast_exception, world
import scipy.linalg as linalg


@pytest.mark.parametrize('dtype', [float, complex])
def test_eighg(dtype):
    n = 5
    S0 = Matrix(n, n, dist=(world, 1, 1), dtype=dtype)
    S = S0.new(dist=(world, world.size, 1))
    H0 = S0.new()
    H = S.new()
    if world.rank == 0:
        S0.data[:] = np.eye(n)
        H0.data[:] = 0.0
        H0.data.ravel()[::n + 1] = np.arange(n) + 1
        if dtype == float:
            S0.data[-1, 0] = 0.001
            H0.data[-1, 0] = 0.001
        else:
            S0.data[-1, 0] = 0.001j
            H0.data[-1, 0] = 0.001j
    H0.tril2full()
    S0.tril2full()
    H00 = H0.copy()
    S0.redist(S)
    H0.redist(H)
    L = S.copy()
    L.invcholesky()
    if world.rank == 0:
        eigs0, C0 = linalg.eigh(H0.data, S0.data)
        print(eigs0)
        print(C0)
        error = H0.data @ C0 - S0.data @ C0 @ np.diag(eigs0)
        print(error)

    L0 = S0.new()
    L.redist(L0)
    eigs = H.eighg(L0)
    H.redist(H0)
    print(eigs)
    print(H0.data)
    with broadcast_exception(world):
        if world.rank == 0:
            assert abs(eigs - eigs0).max() < 1e-14
            error = H00.data @ H0.data - S0.data @ H0.data @ np.diag(eigs)
            assert abs(error).max() < 1e-14



compiler='mpicc'
extra_link_args += [
     '-Wl,--no-as-needed',
    '-L/opt/bwhpc/common/compiler/intel/compxe.2013.sp1.4.211/mkl/lib/intel64',
    '-lmkl_scalapack_lp64',
    '-lmkl_intel_lp64',
    '-lmkl_core',
    '-lmkl_sequential',
    '-lmkl_blacs_intelmpi_lp64',
]



# web-page: pulse.png
import numpy as np
import matplotlib.pyplot as plt
from gpaw.tddft.units import au_to_fs

plt.figure(figsize=(8, 4))
ax = plt.subplot(1, 1, 1)

data_ti = np.loadtxt('dmpulse.dat')
ax.plot(data_ti[:, 0] * au_to_fs, data_ti[:, 2], 'k')
ax.spines['top'].set_visible(False)
ax.yaxis.set_ticks_position('left')
ax.xaxis.set_ticks_position('bottom')
plt.ylabel('Dipole moment (atomic units)')
plt.ylim(np.array([-1, 1]) * np.max(np.abs(plt.ylim())))
plt.xlabel('Time (fs)')
plt.xlim(0, 30)

ax = ax.twinx()
data_ti = np.loadtxt('pulse.dat')
ax.plot(data_ti[:, 0] * au_to_fs, data_ti[:, 1], 'g')
ax.spines['top'].set_visible(False)
ax.yaxis.set_ticks_position('right')
ax.tick_params(axis='y', labelcolor='g')
plt.ylabel('Pulse (atomic units)', color='g')
plt.ylim(np.array([-1, 1]) * np.max(np.abs(plt.ylim())))

plt.tight_layout()
plt.savefig('pulse.png')



from gpaw import GPAW
from ase import Atoms
from gpaw.lrtddft import LrTDDFT

molecule = Atoms('Na2', positions=((0.0, 0.0, 0.0), (3.12, 0.0, 0.0)))

molecule.center(vacuum=4.0)

calc = GPAW(xc='PBE', setups={'Na': '1'}, h=0.25)

molecule.calc = calc

molecule.get_potential_energy()

lr = LrTDDFT(calc, xc='LDA')
lr.write('Omega_Na2.gz')



#!/usr/bin/env python
"""bgp_xlc.py is a wrapper for the BGP xlc compiler,
   converting/removing incompatible gcc args.   """
import sys
from subprocess import call
from glob import glob

args2change = {"-fno-strict-aliasing":"",
               "-fmessage-length=0":"",
               "-Wall":"",
               "-std=c99":"-qlanglvl=extc99",
               "-fPIC":"",
               "-g":"",
               "-D_FORTIFY_SOURCE=2":"",
               "-DNDEBUG":"",
               "-UNDEBUG":"",
               "-pthread":"",
               "-shared":"-qmkshrobj",
               "-Xlinker":"",
               "-export-dynamic":"",
               "-Wstrict-prototypes":"",
               "-dynamic":"",
               "-O3":"",
               "-O2":"",
               "-O1":"",
               "-fwrapv":""}

fragile_files = ["test.c"]
qhot_files = ["c/blas.c", "c/utilities.c","c/lfc.c","c/localized_functions.c"]
non_c99files = glob('c/libxc/src/*.c')

cmd = ""
opt = 1

for arg in sys.argv[1:]:
    cmd += " "
    t = arg.strip()
    if t in fragile_files:
        opt = 2
    if t in non_c99files:
        opt = 3
    if t in qhot_files:
        opt = 4
    if t in args2change:
        cmd += args2change[t]
    else:
        cmd += arg

flags_list = {1: "-g -O3 -qlanglvl=extc99 -qflag=w:w",
              2: "-g -O3 -qlanglvl=extc99 -qflag=w:w",
              3: "-g -O3 -qflag=w:w ",
              4: "-g -O3 -qnohot -qsimd=noauto -qlanglvl=extc99 -qflag=w:w",
              }

flags = flags_list[opt]  
cmd = f"mpixlc_r {flags} {cmd}"

print(f"\nexecmd: {cmd}\n")
call(cmd, shell=True)



from myqueue.workflow import run


def workflow():
    with run(script='cu_calc.py', cores=4, tmax='1h'):
        run(script='cu_plot.py')
        run(function=check)


def check():
    import numpy as np
    from ase.io import read
    energies = []
    k = 20
    for name in ['ITM', 'FD-0.05', 'MV-0.2']:
        e = read(f'Cu-{name}-{k}.txt').get_potential_energy()
        energies.append(e)
    # Extrapolate TM:
    e19 = read('Cu-TM-19.txt').get_potential_energy()
    e20 = read('Cu-TM-20.txt').get_potential_energy()
    e = np.polyval(np.polyfit([20**-2, 19**-2], [e20, e19], 1), 0)
    energies.append(e)
    assert max(energies) - min(energies) < 0.001



import numpy as np
import pytest

import gpaw.tetrahedron as T


@pytest.mark.serial
def test_bja():
    """Test formulas from Blöchl et al. paper."""
    eigs = np.linspace(1.0, 4.0, 4).reshape((4, 1))
    for n in range(1, 4):
        bjan = getattr(T, f'bja{n}')
        bjanb = getattr(T, f'bja{n}b')
        ef = n + 0.5
        E = eigs - ef
        f1, dfde1 = bjan(*E)
        x = 0.0001
        dfde2 = (bjan(*(E - x))[0] - bjan(*(E + x))[0]) / (2 * x)
        f2 = bjanb(*E).sum()
        print(n, f1, f2, dfde1, dfde2)
        assert f1 == pytest.approx(f2)
        assert dfde1 == pytest.approx(dfde2)


@pytest.mark.serial
def test_tetra():
    """Test 2-d BZ exapmle."""
    t = T.TetrahedronMethod(np.diag([1.0, 1.0, 0.1]),
                            [2, 2, 1],
                            False,
                            [0, 1, 2, 1])

    eig_in = np.array([[0.0, 2.0], [0.0, 1.0], [0.0, 1.0]])
    weight_i = [0.25, 0.5, 0.25]
    f_in, (ef,), _ = t.calculate(1.5, eig_in, weight_i, [1.5])
    assert ef == pytest.approx(2 - (2 / 3)**0.5)
    assert f_in.sum(1).dot(weight_i) == pytest.approx(1.5)



from ase import Atom, Atoms
from gpaw import GPAW, FermiDirac
from gpaw.test import equal


def test_spin_spinpol():
    a = 4.0
    n = 16
    hydrogen = Atoms([Atom('H')], cell=(a, a, a), pbc=True)
    hydrogen.center()
    calc = GPAW(gpts=(n, n, n), nbands=1, convergence={'energy': 1e-5},
                occupations=FermiDirac(0.0))
    hydrogen.calc = calc
    e1 = hydrogen.get_potential_energy()
    hydrogen.set_initial_magnetic_moments([1.0])
    e2 = hydrogen.get_potential_energy()
    de = e1 - e2
    print(de)
    equal(de, 0.7871, 1.e-4)

    energy_tolerance = 0.0006
    equal(e1, -0.499854, energy_tolerance)
    equal(e2, -1.287, energy_tolerance)



from ase.build import molecule
from gpaw import GPAW
from gpaw.odd.sic import SIC

n = molecule('N')
n.center(vacuum=3)
calc = GPAW(xc=SIC(nspins=2),
            h=0.2,
            convergence=dict(eigenvalues=1e-5, density=1e-3),
            txt='N.txt')
n.calc = calc
e1 = n.get_potential_energy()

n2 = molecule('N2')
n2.center(vacuum=3)
calc = GPAW(xc=SIC(nspins=1),
            h=0.2,
            convergence=dict(eigenvalues=1e-5, density=1e-3),
            txt='N2.txt')
n2.calc = calc
e2 = n2.get_potential_energy()

assert abs(e2 - 2 * e1 - -4.5) < 0.1



"""Compute the isotropic exchange constants of Co on a band path between
selected high-symmetry points of the hcp lattice."""

# General modules
import numpy as np

# Script modules
from ase.data import covalent_radii

from gpaw import restart
from gpaw.mpi import rank
from gpaw.response.site_kernels import (SphericalSiteKernels,
                                        CylindricalSiteKernels,
                                        ParallelepipedicSiteKernels)
from gpaw.response.chiks import ChiKS
from gpaw.response.mft import IsotropicExchangeCalculator


# ---------- Inputs ---------- #

# Parameters reused from the ground state
gpw = 'Co.gpw'
kpts = 24
nbands = 2 * 6

# Convergence criteria of Kohn-Sham orbitals
conv = {'bands': nbands,
        'forces': 1.e-8,
        'eigenstates': 1.e-14}

# We choose the plane wave energy cutoff of the mft response calculation
# so as to provide magnon energies converged within 5%, based on the
# convergence study of Co(fcc) in [arXiv:2204.04169]
ecut = 750  # eV
# We compute the transverse magnetic susceptibility without broadening
eta = 0.

# We map out the high-symmetry path G-M-K-G-A, by generating all commensurate
# q-vectors on the path
qGM_qc = np.array([[x / kpts, 0., 0.]
                   for x in range(kpts // 2 + 1)])
qMK_qc = np.array([[1 / 2. - x / kpts, 2 * x / kpts, 0.]
                   for x in range(kpts // 6 + 1)])
qKG_qc = np.array([[x / kpts, x / kpts, 0.]
                   for x in reversed(range(kpts // 3 + 1))])
qGA_qc = np.array([[0., 0., x / kpts]
                   for x in range(kpts // 2 + 1)])
q_qc = np.vstack([qGM_qc, qMK_qc[1:], qKG_qc[1:], qGA_qc[1:]])

# We define several Co site radii to try for the spherical site kernels
rc_r = np.linspace(0.5, 1.75, 51)

# ---------- Script ---------- #

# Recalculate the Kohn-Sham orbitals
atoms, calc = restart(gpw, parallel={'domain': 1})
calc.set(fixdensity=True,
         convergence=conv,
         txt='Co_es.txt')
atoms.calc = calc
atoms.get_potential_energy()

# Initialize the ChiKS calculator
chiks = ChiKS(calc,
              ecut=ecut, nbands=nbands, eta=eta,
              gammacentered=True,
              txt='Co_chiks.txt')

# Initialize the exchange calculator
isoexch_calc = IsotropicExchangeCalculator(chiks)

# Initialize site kernels with two sublattices
positions = atoms.positions  # sublattice positions
# Create two Co sublattices with spherical sites, but vary only one radius
rc1_pa = np.array([[rc, 1.2] for rc in rc_r])
sph_sitekernels1 = SphericalSiteKernels(positions, rc1_pa)
assert sph_sitekernels1.nsites == 2  # Check to illustrate SiteKernels magic
assert sph_sitekernels1.npartitions == len(rc_r)
# Create two Co sublattices with spherical sites of equal size
rc2_pa = np.array([[rc, rc] for rc in rc_r])
# We could initialize the site kernel instance normally as follows
sph_sitekernels2 = SphericalSiteKernels(positions, rc2_pa)
# However, there is also a second option. We can initialize a site kernels
# instance for each sublattice and then add them together
sph_sitekernels2 = SphericalSiteKernels(positions[:1], rc2_pa[:, :1])\
    + SphericalSiteKernels(positions[1:], rc2_pa[:, 1:])
assert sph_sitekernels2.nsites == 2
assert sph_sitekernels2.npartitions == len(rc_r)

# Initialize site kernels with a single sublattice
# Create a single site spanning the entire unit cell
cell_cv = atoms.get_cell()
center_v = np.sum(cell_cv, axis=0) / 2.
uc_sitekernels = ParallelepipedicSiteKernels([center_v], [[cell_cv]])
# Create a single site as a cylinder encapsulating both Co atoms
d_v = positions[1] - positions[0]  # displacement vector between atoms
d = np.linalg.norm(d_v)
ez_v = d_v / d  # normalized cylinder axis
# Use the average covalent radii as the cylinder radius
rc = np.average([covalent_radii[n] for n in atoms.get_atomic_numbers()])
# Cylinder height
hc = d + 2 * rc
cyl_sitekernels = CylindricalSiteKernels([center_v], [[ez_v]], [[rc]], [[hc]])
# Because both the unit cell and cylindrical site kernels are placed at the
# same position (the cell center), we may view them as spatial partitionings
# of the same Heisenberg model. Therefore, we can append one site kernel to
# the other, which will create a new "partition".
mix_sitekernels = uc_sitekernels.copy()  # Make a fresh copy
mix_sitekernels.append(cyl_sitekernels)
assert mix_sitekernels.nsites == 1
assert mix_sitekernels.npartitions == 2

# Allocate arrays for the exchange constants
nq = len(q_qc)
Jsph1_qabr = np.empty((nq, 2, 2, len(rc_r)), dtype=complex)
Jsph2_qabr = np.empty((nq, 2, 2, len(rc_r)), dtype=complex)
Jmix_qp = np.empty((nq, 2), dtype=complex)

# Compute the isotropic exchange coupling along the chosen bandpath
for q, q_c in enumerate(q_qc):
    Jsph1_qabr[q] = isoexch_calc(q_c, sph_sitekernels1)
    Jsph2_qabr[q] = isoexch_calc(q_c, sph_sitekernels2)
    Jmix_qp[q] = isoexch_calc(q_c, mix_sitekernels)[0, 0, :]  # nsites == 1

# Save the bandpath, spherical radii and computed exchange constants
if rank == 0:
    np.save('Co_q_qc.npy', q_qc)
    np.save('Co_rc_r.npy', rc_r)
    np.save('Co_Jsph1_qabr.npy', Jsph1_qabr)
    np.save('Co_Jsph2_qabr.npy', Jsph2_qabr)
    np.save('Co_Jmix_qp.npy', Jmix_qp)



from gpaw import GPAW

energies = []
for name in ['zero', 'periodic', 'corrected']:
    calc = GPAW(name + '.gpw', txt=None)
    energies.append(calc.get_potential_energy())

print(energies)
assert abs(energies[1] - energies[0]) < 0.003
assert abs(energies[2] - energies[0] - 0.0409) < 0.003

efermi = calc.get_fermi_level()

v = calc.get_electrostatic_potential().mean(1).mean(0)
w1 = v[0] - efermi
w2 = v[-1] - efermi
print(w1, w2)
assert abs(w1 - 4.359) < 0.01
assert abs(w2 - 2.556) < 0.01



# This script exercsises some of the idiosyncracies
# of the descriptor class and PBLAS on a realistic
# case. See the BLACS descriptor documentation
# in trunk/gpaw/blacs.py for some discussions of
# these idiosyncracies.
import numpy as np

from gpaw.blacs import BlacsGrid, parallelprint
from gpaw.mpi import world
from gpaw.utilities.scalapack import pblas_simple_gemm

gen = np.random.RandomState(42)

# simulate state-parallelization=2 and
# domain-decomposition.prod=32
B = 2
D = 32
mb = 32
grid = BlacsGrid(world, B, D)

nbands = 500
nG = 80**3

nGdesc = grid.new_descriptor(nbands, nG, nbands // B, nG // D)
nndesc = grid.new_descriptor(nbands, nbands, mb, mb)

psit_nG = gen.rand(*nGdesc.shape)
A_nn = gen.rand(*nndesc.shape)

assert nGdesc.check(psit_nG)
assert nndesc.check(A_nn)

parallelprint(world, (A_nn.shape, nndesc.shape, nndesc.lld))

pblas_simple_gemm(nGdesc, nGdesc, nndesc, psit_nG, psit_nG, A_nn,
                  transa='N', transb='T')



"""Test ITM with half-filled 1-d cosine band."""
from math import pi
import pytest
import numpy as np
from gpaw.tetrahedron import TetrahedronMethod
from gpaw.dos import DOSCalculator


def f(N: int, i: bool):
    rcell = np.diag([1, 0.1, 0.1])
    k = (np.linspace(-0.5, 0.5, N, endpoint=False) + 0.5 / N) * 2 * pi
    e = -np.cos(k)[:, np.newaxis]
    f = np.empty((N, 1))
    w = np.zeros(N) + 1 / N
    t = TetrahedronMethod(rcell, (N, 1, 1), improved=i)
    ef, _ = t._calculate(0.5, e, w, f)
    wfs = WFS(e, ef)
    dos = DOSCalculator(wfs, cell=np.linalg.inv(rcell))
    dosef = dos.raw_dos([ef], spin=0, width=0.0)[0]
    return f.sum() / N, (f * e).sum() / N, ef, dosef


class WFS:
    def __init__(self, e, ef):
        self.eig_skn = e[np.newaxis]
        self.fermi_level = ef
        self.size = (len(e), 1, 1)
        self.bz2ibz_map = None

    def weights(self):
        n = self.size[0]
        return np.zeros(n) + 1 / n

    def eigenvalues(self):
        return self.eig_skn


@pytest.mark.serial
def test_tm_1d():
    # Improved:
    nelectrons, eband, efermi, dosefermi = f(100, True)
    assert nelectrons == pytest.approx(0.5, abs=1e-9)
    assert eband == pytest.approx(-1 / pi, abs=1e-5)
    assert efermi == pytest.approx(0.0, abs=1e-9)
    assert dosefermi == pytest.approx(1 / pi, abs=1e-4)

    # Standard:
    nelectrons, eband100, efermi, dosefermi = f(100, not True)
    assert nelectrons == pytest.approx(0.5, abs=1e-9)
    assert efermi == pytest.approx(0.0, abs=1e-9)
    assert dosefermi == pytest.approx(1 / pi, abs=1e-4)

    # ebands converges as 1/N^2.
    # Extrapolate to infinite number of k-points:
    eband80 = f(80, not True)[1]
    eband = np.polyval(np.polyfit([100**-2, 80**-2],
                                  [eband100, eband80], 1), 0.0)
    assert eband == pytest.approx(-1 / pi, abs=1e-7)



import pytest
import numpy as np
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
# from ase.units import Bohr

from gpaw import GPAW, FermiDirac
from gpaw.response.chi0 import Chi0
from gpaw.mpi import serial_comm


@pytest.mark.response
@pytest.mark.slow
def test_response_chi0(in_tmp_dir):
    omega = np.array([0, 1.0, 2.0])
    for k in [2, 3]:
        q_c = [0, 0, 1.0 / k]
        for gamma in [False, True]:
            if k == 3 and gamma:
                continue
            kpts = monkhorst_pack((k, k, k))
            if gamma:
                kpts += 0.5 / k
            for center in [False, True]:
                a = bulk('Si', 'diamond')
                if center:
                    a.center()
                for sym in [False, True]:
                    name = 'si.k%d.g%d.c%d.s%d' % (k, gamma, center, bool(sym))
                    print(name)
                    if 1:
                        calc = a.calc = GPAW(
                            kpts=kpts,
                            eigensolver='rmm-diis',
                            symmetry={'point_group': sym},
                            mode='pw',
                            occupations=FermiDirac(width=0.001),
                            txt=name + '.txt')
                        a.get_potential_energy()
                        calc.write(name, 'all')

                    calc = GPAW(name, txt=None, communicator=serial_comm)

                    chi = Chi0(calc, frequencies=omega, hilbert=False,
                               ecut=100, txt=name + '.log')
                    chi0 = chi.calculate(q_c)
                    chi0_wGG = chi0.chi0_wGG

                    if not sym and not center:
                        chi00_w = chi0_wGG[:, 0, 0]
                    elif -1 not in calc.wfs.kd.bz2bz_ks:
                        assert abs(chi0_wGG[:, 0, 0] - chi00_w).max() < 3e-5

                    if not sym:
                        chi00_wGG = chi0_wGG
                    elif -1 not in calc.wfs.kd.bz2bz_ks:
                        assert abs(chi0_wGG - chi00_wGG).max() < 2e-5

                    # q0_c = [0, 1e-7, 1e-7]
                    # q0_v = np.dot(q0_c,
                    #               a.get_reciprocal_cell() * 2 * np.pi) * Bohr
                    # q0 = (q0_v**2).sum()**0.5

                    chi0 = chi.calculate([0, 0, 0])
                    chi0_wGG = chi0.chi0_wGG

                    if not sym and not center:
                        chi000_w = chi0_wGG[:, 0, 0]
                    elif -1 not in calc.wfs.kd.bz2bz_ks:
                        assert abs(chi0_wGG[:, 0, 0] - chi000_w).max() < 0.001

                    if not sym:
                        chi000_wGG = chi0_wGG
                    elif -1 not in calc.wfs.kd.bz2bz_ks:
                        assert abs(chi0_wGG - chi000_wGG).max() < 0.001



from gpaw import GPAW, restart
import pickle

slab, calc = restart('top.gpw')
c_mol = GPAW('CO.gpw')
molecule = range(len(slab))[-2:]
e_n = []
P_n = []
for n in range(c_mol.get_number_of_bands()):
    print('Band: ', n)
    wf_k = [kpt.psit_nG[n] for kpt in c_mol.wfs.kpt_u]
    P_aui = [[kpt.P_ani[a][n] for kpt in c_mol.wfs.kpt_u]
             for a in range(len(molecule))]
    e, P = calc.get_all_electron_ldos(mol=molecule, wf_k=wf_k, spin=0,
                                      P_aui=P_aui, raw=True)
    e_n.append(e)
    P_n.append(P)
pickle.dump((e_n, P_n), open('top.pickle', 'wb'))



import matplotlib.pyplot as plt
import numpy as np
from gpaw import GPAW
from gpaw.response.df import DielectricFunction

calc = GPAW('Ag_LDA.gpw', parallel={'domain': 1})
calc.diagonalize_full_hamiltonian(nbands=30)
calc.write('Ag_LDA_full.gpw', 'all')

# Set up dielectric function:
df = DielectricFunction(calc='Ag_LDA_full.gpw',  # Ground state input
                        domega0=0.05)  # energy grid spacing at omega=0

# Momentum transfer, must be the difference between two kpoints!
q_c = [1.0 / 10, 0, 0]
df.get_eels_spectrum(q_c=q_c)  # a file called 'eels.csv' is generated


# Plot spectrum
data = np.loadtxt('eels.csv', delimiter=',')
omega = data[:, 0]
eels = data[:, 2]
plt.plot(omega, eels)
plt.xlabel('Energy (eV)')
plt.ylabel('Loss spectrum')
plt.xlim(0, 20)
plt.show()



import numpy as np
from ase.dft.kpoints import monkhorst_pack
from ase.parallel import paropen
from ase.build import hcp0001, add_adsorbate
from gpaw import GPAW, PW, FermiDirac
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy

kpts = monkhorst_pack((16, 16, 1))
kpts += np.array([1 / 32, 1 / 32, 0])

a = 2.51  # lattice parameter of Co
slab = hcp0001('Co', a=a, c=4.07, size=(1, 1, 4))
pos = slab.get_positions()
cell = slab.get_cell()
cell[2, 2] = 20. + pos[-1, 2]
slab.set_cell(cell)
slab.set_initial_magnetic_moments([0.7, 0.7, 0.7, 0.7])

ds = [1.75, 2.0, 2.25, 2.5, 2.75, 3.0, 3.25, 3.5, 3.75, 4.0, 5.0, 6.0, 10.0]

for d in ds:
    pos = slab.get_positions()
    add_adsorbate(slab, 'C', d, position=(pos[3, 0], pos[3, 1]))
    add_adsorbate(slab, 'C', d, position=(cell[0, 0] / 3 + cell[1, 0] / 3,
                                          cell[0, 1] / 3 + cell[1, 1] / 3))
    # view(slab)
    calc = GPAW(xc='PBE',
                mode=PW(600),
                kpts=kpts,
                occupations=FermiDirac(width=0.01),
                txt=f'hmm2_gs_{d}.txt')
    slab.calc = calc
    E = slab.get_potential_energy()
    E_hf = nsc_energy(calc, 'EXX')

    calc.diagonalize_full_hamiltonian()
    calc.write(f'gs_{d}.gpw', mode='all')

    f = paropen('hf_acdf.dat', 'a')
    print(d, E, E_hf, file=f)
    f.close()

    del slab[-2:]



from ase.build import mx2
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(400),
            xc='PBE',
            nbands=70,
            convergence={'bands': -20},
            setups={'Mo': '6'},
            parallel={'band': 1, 'domain': 1},
            occupations=FermiDirac(width=0.001),
            kpts={'size': (42, 42, 1), 'gamma': True},
            txt='gs_MoS2.txt')

slab = mx2(formula='MoS2', a=3.16, thickness=3.17, vacuum=5.0)
slab.calc = calc
slab.get_potential_energy()

calc.write('gs_MoS2.gpw', mode='all')



import numpy as np
from gpaw import restart

slab, calc = restart('ontop.gpw', txt=None)
AuAl_density = calc.get_pseudo_density()

# Remove gold atom and do a clean slab calculation:
del slab[-1]
slab.get_potential_energy()
Al_density = calc.get_pseudo_density()

# Remove Al atoms and do a calculation for Au only:
slab, calc = restart('ontop.gpw', txt=None)
del slab[:-1]
calc.set(kpts=None)
slab.get_potential_energy()
Au_density = calc.get_pseudo_density()

diff = AuAl_density - Au_density - Al_density
np.save('densitydiff.npy', diff)



# web-page: Bi2Se3_bands.png
import numpy as np
import matplotlib.pyplot as plt

from gpaw import GPAW
from gpaw.spinorbit import soc_eigenstates


calc = GPAW('Bi2Se3_bands.gpw', txt=None)
x = np.loadtxt('kpath.dat')
X = np.loadtxt('highsym.dat')

# No spin-orbit

ef = calc.get_fermi_level()
e_kn = np.array([calc.get_eigenvalues(kpt=k)
                 for k in range(len(calc.get_ibz_k_points()))])
e_nk = e_kn.T
e_nk -= ef

for e_k in e_nk:
    plt.plot(x, e_k, '--', c='0.5')

# Spin-orbit calculation
soc = soc_eigenstates(calc, scale=1.0)
e_nk = soc.eigenvalues().T - soc.fermi_level

plt.xticks(X, [r'$\Gamma$', 'Z', 'F', r'$\Gamma$', 'L'], size=24)
plt.yticks(size=20)
for i in range(len(X))[1:-1]:
    plt.plot(2 * [X[i]], [1.1 * np.min(e_nk), 1.1 * np.max(e_nk)],
             c='0.5', linewidth=0.5)
for e_k in e_nk:
    plt.plot(x, e_k, c='b')

plt.ylabel(r'$\varepsilon_n(k)$ [eV]', size=24)
plt.axis([0, x[-1], -1.7, 1.7])
plt.tight_layout()
# plt.show()
plt.savefig('Bi2Se3_bands.png')



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""

# Noncollinear magnetism - VI$_2$

For this part of the project, we will move on to a magnetic monolayer in which the dominant exchange coupling is antiferromagnetic, namely VI$_2$. We will still use the localized spin Hamiltonian

$$H = -\frac{1}{2}\sum_{i,j}J_{ij}\mathbf{S}_i\cdot \mathbf{S}_j+A\sum_i(S_i^z)^2$$

but here in the antiferromagnetic case, $J<0$.

## Optimizing the atomic structure

Since we will need to do LDA calculations later on, we will start of this part of the project by relaxing the atomic structure of VI$_2$ using the LDA functional. Usually, the difference in crystal structure between different magnetically ordered states is small, so we will perform the relaxation in the ferromagnetic state, which has a smaller unit cell.

1.   First you should download the relaxed PBE crystal structure. Either, browse the C2DB at https://cmrdb.fysik.dtu.dk/c2db and download the `.xyz` file for VI$_2$ or dowload it directly from the summer school tutorial website [here](https://wiki.fysik.dtu.dk/gpaw/summerschools/summerschool22/magnetism/magnetism.html)
2.   Fill in the expected ionic value for the V spins `S` below and run the cell to relax the crystal structure. The calculation takes about 17 minutes. (Hint: V has the electronic configuration [Ar]3d$^3$4s$^2$)

"""

# %%
from ase.io import read
from ase.visualize import view
from ase.optimize import BFGS
from ase.constraints import UnitCellFilter
from gpaw import GPAW, PW

S = 3 / 2  # student: S = ???
m = S * 2
layer = read('VI2.xyz')
layer.set_initial_magnetic_moments([m, 0, 0])
view(layer)

calc = GPAW(mode=PW(400), xc='LDA', kpts=(4, 4, 1))
layer.calc = calc

uf = UnitCellFilter(layer, mask=[1, 1, 0, 0, 0, 1])
opt = BFGS(uf)
opt.run(fmax=0.1)

calc.set(symmetry='off')
calc.get_potential_energy()

calc.write('VI2_relaxed.gpw')

# %%
"""
## Magnetic anisotropy

In the cell above, we not only performed a structural optimization, but we also took the relaxed structure, switched off the $k$-point symmetries and did a final DFT calculation with all the $k$-points in the Brillouin zone. We did this in order to be able to evaluate the the magnetic anisotropy arising from the spin-orbit coupling.

1.   Adapt the code you used for CrI$_3$ to calculate the magnetic anisotropy in the cell below.
2.   Is the easy-axis in-plane or out-of-plane?
3.   Do you expect to the VI$_2$ to exhibit magnetic order at finite temperatures?

"""

# %%
from gpaw.spinorbit import soc_eigenstates
# teacher:

calc = GPAW('VI2_relaxed.gpw')
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy()
    for theta, phi in [(90, 0), (90, 90), (0, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'dE_zx = {de_zx * 1000:1.3f} meV')
print(f'dE_zy = {de_zy * 1000:1.3f} meV')
A = (de_zx + de_zy) / 2 / S**2
print(f'A = {A * 1000:1.3f} meV')

# %%
"""
## DFT calculations in a repeated cell

To realize that VI$_2$ is in fact an antiferromagnetic material, we need to do a calculation starting from an antiferromagnetic alignment of the spins. To do so, we need more than a single V atom in the unit cell. In the cell below, the atomic structure is repeated once to obtain two V atoms in the unit cell and it is shown how do a DFT calculation for the antiferromagnetic state.

1.   Fill in the `kpts` for the repeated unit cell.
2.   Replace the `...` with code to calculate the ferromagnetic state.
3.   When you have finalzed the code for the ferromagnetic state, run the cell. The calculation takes about 7 minutes.

"""

# %%
layer_afm = layer.repeat((2, 1, 1))
layer_afm.set_initial_magnetic_moments([m, 0, 0, -m, 0, 0])
view(layer_afm)

calc = GPAW(mode=PW(400),
            xc='LDA',
            kpts=(2, 4, 1))  # student: kpts=???))
layer_afm.calc = calc
layer_afm.get_potential_energy()
calc.write('V2I4_afm.gpw')

...
# teacher:
layer_fm = layer.repeat((2, 1, 1))
layer_fm.set_initial_magnetic_moments([m, 0, 0, m, 0, 0])
calc = GPAW(mode=PW(400), xc='LDA', kpts=(2, 4, 1))
layer_fm.calc = calc
layer_fm.get_potential_energy()
calc.write('V2I4_fm.gpw')

# %%
"""
## Calculating J

Finally, we are in a position to calculate the nearest neighbour Heisenberg exchange coupling $J$. Before doing so, please compare the output of the antiferromagnetic and ferromagnetic calculations from the cell above.

1.   Which state has the lowest energy?
2.   What will the sign of $J$ be?

You should find that the antiferromagnetic state has a lower energy than the ferromagnetic one, but does this also mean that the calculated configuration is the correct magnetic ground state? Rather, it implies that the spins prefer to be antialigned, i.e. that the exchange coupling $J$ is negative.

3.   Draw the structural arrangement of the V atoms on a piece of paper. Which type of magnetic lattice do they form?
4.   Fill in the spin configuration of the magnetic lattice and convince yourself that all spins cannot be antialigned to their nearest neighbours.

The latter finding means that the antiferromagnetic system is frustrated and the antiferromagnetic configuration we have computed will not be the true ground state of the system.

Leaving the magnetic frustration aside for the moment, we will first calculate $J$, which we can still do even though that we have not found the ground state yet.

5.   Use the classical Heisenberg model with nearest neighbor interaction only (and $A=0$) to derive the energy per magnetic site of the ferromagnetic and antiferromagnetic configurations respectively.

You should obtain the following:

$$E_{\mathrm{FM}} = E_0 - 3JS^2$$

and

$$E_{\mathrm{AFM}} = E_0 + JS^2$$

where $E_0$ is some reference energy.

6.   Use these expressions to eliminate $E_0$ and express $J$ in terms of the energy difference per magnetic site of the two configurations.
7.   Write code to extract `E_fm` and `E_afm` in the cell below.
8.   Fill in the formula for `J` and evaluate the cell to calculate it.

You should get a value for $J$ around -1.4 meV.
"""

# %%
E_fm = layer_fm.get_potential_energy() / 2  # student: E_fm = ???
E_afm = layer_afm.get_potential_energy() / 2  # student: E_afm = ???
dE = E_afm - E_fm  # student:
J = dE / 4 / S**2  # student: J = ???
print(f'J = {J * 1000:1.2f} meV')

# %%
"""
## Noncollinear configuration

As it turn out, the optimal spin structure on a trigonal lattice with antiferromagntice exchange coupling is to place all spins at 120$^\circ$ angles with respect its neighbors.

1.   Draw this structure and convince yourself that it is indeed possible to put every spin at a 120$^\circ$ angle with respect to its neighbors.
2.   What is the minimal number of magnetic atoms required in the magnetic unit cell to represent such a state?
3.   Verrify that the Heisenberg model with classical spins gives a lower energy with this configuration than in the antialigned structure calculated above.

You should obtain the following energy for the 120$^\circ$ noncolinnear configuration:

$$E_{\mathrm{NC}}=E_0+\frac{3}{2}JS^2.$$

We will now check if we can verify this prediction within the LSDA. To do that we need to perform a noncollinear DFT calculation, which is done in the cell below.

4.   Read and try to understand the code to perform a noncollinear LSDA calculation in the 120$^\circ$ noncolinnear configuration.
5.   Replace the `...` with code to make a noncolinnear LSDA calculation for the ferromagnetic state as well (we will need this for a late comparison).

Run the cell and verify that the energy per magnetic atom is lower in the 120$^\circ$ noncolinnear configuration compared to both of the previous calculated states.
"""

# %%
import numpy as np
from ase.io import read
from ase.visualize import view
from gpaw import GPAW, PW, MixerDif

m = 3
cell_cv = layer.get_cell()
layer_nc = layer.repeat((3, 1, 1))
new_cell_cv = [[3 * cell_cv[0, 0] / 2, 3**0.5 * cell_cv[0, 0] / 2, 0.0],
               [3 * cell_cv[0, 0] / 2, -3**0.5 * cell_cv[0, 0] / 2, 0.0],
               [0.0, 0.0, cell_cv[2, 2]]]
layer_nc.set_cell(new_cell_cv)
view(layer_nc)

magmoms = np.zeros((len(layer_nc), 3), float)
magmoms[0] = [m, 0, 0]
magmoms[3] = [m * np.cos(2 * np.pi / 3), m * np.sin(2 * np.pi / 3), 0]
magmoms[6] = [m * np.cos(2 * np.pi / 3), -m * np.sin(2 * np.pi / 3), 0]

calc = GPAW(mode=PW(400),
            xc='LDA',
            mixer=MixerDif(),
            symmetry='off',
            experimental={'magmoms': magmoms, 'soc': False},
            parallel={'domain': 1, 'band': 1},
            kpts=(2, 2, 1))
layer_nc.calc = calc
layer_nc.get_potential_energy()
calc.write('nc_nosoc.gpw')

...
# teacher:
magmoms = np.zeros((len(layer_nc), 3), float)
magmoms[0] = [m, 0, 0]
magmoms[3] = [m, 0, 0]
magmoms[6] = [m, 0, 0]
calc = GPAW(mode=PW(400),
            xc='LDA',
            mixer=MixerDif(),
            symmetry='off',
            experimental={'magmoms': magmoms, 'soc': False},
            parallel={'domain': 1, 'band': 1},
            kpts=(2, 2, 1))
layer_nc.calc = calc
layer_nc.get_potential_energy()
calc.write('fm_nosoc.gpw')

# %%
"""
## Anisotropy and exchange coupling from noncollinear DFT

In the cell above we could have included spin-orbit coupling in a self-consistent way by setting 'soc'=True'. However, it is more convenient for us to exclude it such that we can explicitly compute the single-ion anisotropy parameter $A$ afterwards. If the localized spin Hamiltonian with nearest neighbor exchange interactions is a good model, we should be able to obtain both $J$ and $A$ from the noncollinear calculation as well.

1.   Fill in the cell below with code to compute the nearest neighbour exchange coupling $J$ based on the noncollinear calculations performed above.
2.   Fill in code that computes the spin-orbit coupling corrected energies of the ferromagnetic state from the noncollinear calculation with spins directed along the $x$, $y$ and $z$ directions (Hint: the anisotropy is calculated by rotating the entire initial spin configuration first by $\theta$ and then by $\varphi$).
3.   Repeat point 2, but for the 120$^\circ$ noncolinnear configuration. In this case, you cannot align all spins to one direction, but takes as a reference the first V atom.

With the above code in place, please evaluate the cell. You should obtain a nearest neighbour exchange coupling of about -1.9 meV.

"""

# %%

# teacher:
calc = GPAW('nc_nosoc.gpw', txt=None)
E_nc = calc.get_potential_energy() / 3
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy() / 3
    for theta, phi in [(0, 0), (0, 90), (90, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'NC: A_zx = {de_zx * 1000:1.3f} meV')
print(f'NC: A_zy = {de_zy * 1000:1.3f} meV')
print()

calc = GPAW('fm_nosoc.gpw', txt=None)
E_fm = calc.get_potential_energy() / 3
e_x, e_y, e_z = (
    soc_eigenstates(calc, theta=theta, phi=phi).calculate_band_energy() / 3
    for theta, phi in [(0, 0), (0, 90), (90, 0)])
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'FM: A_zx = {de_zx * 1000:1.3f} meV')
print(f'FM: A_zy = {de_zy * 1000:1.3f} meV')
print()

dE = E_nc - E_fm
J = 2 * dE / 9 / S**2
print(f'J = {J * 1000:1.2f} meV')
print()
calc = GPAW('VI2_relaxed.gpw', txt=None)
E_fm = calc.get_potential_energy()
dE = E_nc - E_fm
J = 2 * dE / 9 / S**2
print(f'J = {J * 1000:1.2f} meV')


# %%
"""
## Critical temperature?

Based on the noncollinear calculations above:

1.   What is the easy axis of the system?
2.   Does the easy axis agree with your initial findings for the simple ferromagnetic state?

You should find that the energy of the 120$^\circ$ noncolinnear configuration does not depend strongly on whether the first V atom is aligned to the $x$ or the $y$ direction.

3.   If we assume full in-plane isotropy is there any rotational freedom left in the noncollinear ground state?
4.   What implications does this have for the critical temperature of the monolayer?

You might be able to convince yourself that some degree of in-plane anisotropy is required to obtain a finite critical temperature for the 120$^\circ$ noncolinnear magnetic order. Again, bear in mind that all the calculations in the present notebook ought to be properly converged with respect to $k$-points, plane wave cutoff etc. to achieve an accurate estimate of e.g. the in-plane anisotropy.

Clearly the noncollinear spin state of VI$_2$ is more difficult to describe than the ferromagnetic state in CrI$_3$ and we do not yet have a simple theoretical expression for the critical temperature as a function of anisotropy and exchange coupling constants. However, with the rapid development of experimental techniques to synthesize and characterize 2D materials it does seem plausible that such a noncollinear 2D material may be observed in the future.
"""



"""Calculations using RSF with dedicated datasets and Poisson-solver."""
from ase import Atoms
from gpaw import GPAW, setup_paths
from gpaw.poisson import PoissonSolver
from gpaw.eigensolvers import RMMDIIS
from gpaw.occupations import FermiDirac
from gpaw.test import gen

if setup_paths[0] != '.':
    setup_paths.insert(0, '.')

for atom in ['C', 'O']:
    gen(atom, xcname='PBE', scalarrel=True, exx=True,
        yukawa_gamma=0.75)

h = 0.30
co = Atoms('CO', positions=[(0, 0, 0), (0, 0, 1.15)])
co.center(5)

# c = {'energy': 0.005, 'eigenstates': 1e-4}  # Usable values
c = {'energy': 0.1, 'eigenstates': 3, 'density': 3}  # Values for test

calc = GPAW(txt='CO.txt', xc='LCY-PBE', convergence=c,
            eigensolver=RMMDIIS(), h=h,
            poissonsolver=PoissonSolver(use_charge_center=True),
            occupations=FermiDirac(width=0.0), spinpol=False)
co.calc = calc
co.get_potential_energy()



import pytest
import numpy as np
from ase import Atoms
from gpaw import GPAW, FermiDirac, PW
from gpaw.response.df import DielectricFunction
from gpaw.test import equal, findpeak


@pytest.mark.response
@pytest.mark.slow
def test_response_au02_absorption(scalapack, in_tmp_dir):
    GS = 1
    ABS = 1
    if GS:
        cluster = Atoms('Au2', [(0, 0, 0), (0, 0, 2.564)])
        cluster.set_cell((6, 6, 6), scale_atoms=False)
        cluster.center()
        calc = GPAW(mode=PW(ecut=180, force_complex_dtype=True),
                    xc={'name': 'RPBE', 'stencil': 1},
                    nbands=16,
                    eigensolver='rmm-diis',
                    parallel={'domain': 1},
                    occupations=FermiDirac(0.01))

        cluster.calc = calc
        cluster.get_potential_energy()
        calc.diagonalize_full_hamiltonian(nbands=24, scalapack=True)
        calc.write('Au2.gpw', 'all')

    nw = 141
    frequencies = np.linspace(0, 14, nw)

    if ABS:
        df = DielectricFunction('Au2.gpw',
                                frequencies=frequencies,
                                hilbert=not True,
                                eta=0.1,
                                ecut=10)

        b0, b = df.get_dielectric_function(filename=None,
                                           direction='z')
        a0, a = df.get_polarizability(filename=None,
                                      direction='z')
        df_ws = DielectricFunction('Au2.gpw',
                                   frequencies=frequencies,
                                   hilbert=not True,
                                   eta=0.1,
                                   ecut=10,
                                   truncation='wigner-seitz')

        a0_ws, a_ws = df_ws.get_polarizability(filename=None,
                                               direction='z')

        w0_ = 5.60491055
        I0_ = 227.23392824591642
        w_ = 5.644900254787107
        I_ = 184.4086028397282

        w, I = findpeak(frequencies, b0.imag)
        equal(w, w0_, 0.05)
        equal(6**3 * I / (4 * np.pi), I0_, 0.5)
        w, I = findpeak(frequencies, a0.imag)
        equal(w, w0_, 0.05)
        equal(I, I0_, 0.5)
        w, I = findpeak(frequencies, a0_ws.imag)
        equal(w, w0_, 0.05)
        equal(I, I0_, 0.5)
        w, I = findpeak(frequencies, b.imag)
        equal(w, w_, 0.05)
        equal(6**3 * I / (4 * np.pi), I_, 0.5)
        w, I = findpeak(frequencies, a.imag)
        equal(w, w_, 0.05)
        equal(I, I_, 0.5)
        # The Wigner-Seitz truncation does not give exactly the
        # same for small cell
        w, I = findpeak(frequencies, a_ws.imag)
        equal(w, w_, 0.2)
        equal(I, I_, 8.0)



from ase.units import _amu, _me, Bohr, AUT, Hartree
from gpaw.tddft import TDDFT
from gpaw.tddft.ehrenfest import EhrenfestVelocityVerlet
from ase.io import Trajectory
import numpy as np

name = 'graphene_h'
Ekin = 40e3  # kinetic energy of the ion (in eV)

# Adapted to the ion energy; here 4 as (probably too large!)
timestep = 16.0 * np.sqrt(10e3 / Ekin)
ekin_str = '_ek' + str(int(Ekin / 1000)) + 'k'
strbody = name + ekin_str
traj_file = strbody + '.traj'

# The parallelization options should match the number of cores, here 32.
p_bands = 2  # number of bands to parallelise over
dom_dc = (2, 2, 4)  # domain decomposition for parallelization
parallel = {'band': p_bands, 'domain': dom_dc}

tdcalc = TDDFT(name + '.gpw',
               propagator='EFSICN',
               solver='BiCGStab',
               txt=strbody + '_td.txt',
               parallel=parallel)

proj_idx = 50  # atomic index of the projectile
delta_stop = 5.0 / Bohr  # stop condition when ion is within 5 A of boundary.

# Setting the initial velocity according to the kinetic energy.
amu_to_aumass = _amu / _me
Mproj = tdcalc.atoms.get_masses()[proj_idx] * amu_to_aumass
Ekin *= 1 / Hartree
v = np.zeros((proj_idx + 1, 3))
v[proj_idx, 2] = -np.sqrt((2 * Ekin) / Mproj) * Bohr / AUT
tdcalc.atoms.set_velocities(v)

evv = EhrenfestVelocityVerlet(tdcalc)
traj = Trajectory(traj_file, 'w', tdcalc.get_atoms())

trajdiv = 1  # number of timesteps between trajectory images
densdiv = 10  # number of timesteps between saved electron densities
niters = 100  # total number of timesteps to propagate

for i in range(niters):
    # Stopping condition when projectile z coordinate passes threshold
    if evv.x[proj_idx, 2] < delta_stop:
        tdcalc.write(strbody + '_end.gpw', mode='all')
        break

    # Saving trajectory file every trajdiv timesteps
    if i % trajdiv == 0:
        F_av = evv.F * Hartree / Bohr  # forces converted from atomic units
        v_av = evv.v * Bohr / AUT  # velocities converted from atomic units
        epot = tdcalc.get_td_energy() * Hartree  # energy
        atoms = tdcalc.get_atoms().copy()
        atoms.set_velocities(v_av)

        traj.write(atoms, energy=epot, forces=F_av)

    # Saving electron density every densdiv timesteps
    if (i != 0 and i % densdiv == 0):
        tdcalc.write(strbody + '_step' + str(i) + '.gpw')

    evv.propagate(timestep)

traj.close()



from ase.phonons import Phonons
from gpaw import GPAW
from gpaw.raman.elph import EPC

# Load pre-computed ground state calculation (primitive cell)
calc = GPAW('gs.gpw', parallel={'band': 1})
atoms = calc.atoms

# Load results from phonon and electron-phonon coupling calculations
phonon = Phonons(atoms, supercell=(1, 1, 1))
elph = EPC(atoms, supercell=(2, 2, 2))

# Construct electron-phonon matrix of Bloch functions
elph.get_elph_matrix(calc, phonon)



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw import PW

bulk_c = bulk('C', a=3.5454859)
calc = GPAW(mode=PW(600.0),
            parallel={'domain': 1},
            xc='LDA',
            occupations=FermiDirac(width=0.01),
            kpts={'size': (6, 6, 6), 'gamma': True},
            txt='diam_kern.ralda_01_lda.txt')

bulk_c.calc = calc
E_lda = bulk_c.get_potential_energy()
calc.diagonalize_full_hamiltonian()
calc.write('diam_kern.ralda.lda_wfcs.gpw', mode='all')



import numpy as np

results = np.loadtxt('range_results.dat')

rc_05 = results[1, 1]
rc_10 = results[2, 1]
rc_20 = results[3, 1]
rc_30 = results[4, 1]
rc_40 = results[5, 1]

assert abs(rc_05 - -12.11) < 0.01
assert abs(rc_10 - -12.04) < 0.01
assert abs(rc_20 - -12.22) < 0.01
assert abs(rc_30 - -12.55) < 0.01
assert abs(rc_40 - -12.84) < 0.01



"""Test CLI-dos command."""
import pytest
from gpaw.cli.main import main


@pytest.mark.serial
@pytest.mark.parametrize(
    'options',
    ['',
     '-t',
     '-i',
     '-a Li-sp',
     '-w 0',
     '-r -5 5',
     '-n 50',
     '--soc'])
def test_dos(gpw_files, options):
    args = ('dos ' + str(gpw_files['bcc_li_pw']) + ' ' + options).split()
    main(args)



# web-page: tc_000_with_08.00eV.txt
# Start
from ase.parallel import paropen
from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2

# Maximum energy difference for Kohn-Sham transitions
# included in the calculation
max_energy_diff = 8.0  # eV

calc = GPAW('unocc.gpw')
lr = LrTDDFT2('lr2', calc,
              fxc='LDA',
              max_energy_diff=max_energy_diff,
              txt='-')

# Get and write spectrum
spec = lr.get_spectrum('spectrum_with_%05.2feV.dat'
                       % max_energy_diff,
                       min_energy=0,
                       max_energy=10,
                       energy_step=0.01,
                       width=0.1)

# Get and write transitions
trans = lr.get_transitions('transitions_with_%05.2feV.dat'
                           % max_energy_diff,
                           min_energy=0.0,
                           max_energy=10.0)

# Get and write transition contributions
index = 0
f2 = lr.get_transition_contributions(index_of_transition=index)
with paropen('tc_%03d_with_%05.2feV.txt'
             % (index, max_energy_diff), 'w') as f:
    f.write('Transition %d at %.2f eV\n' % (index, trans[0][index]))
    f.write(' %5s => %5s  contribution\n' % ('occ', 'unocc'))
    for (ip, val) in enumerate(f2):
        if (val > 1e-3):
            f.write(' %5d => %5d  %8.4f%%\n' %
                    (lr.ks_singles.kss_list[ip].occ_ind,
                     lr.ks_singles.kss_list[ip].unocc_ind,
                     val / 2. * 100))



from ase import Atoms
from ase.visualize import view
from gpaw import GPAW
from gpaw.wannier import calculate_overlaps

calc = GPAW('si.gpw')
atoms = calc.get_atoms()

overlaps = calculate_overlaps(calc, nwannier=16)
wan = overlaps.localize_er()

watoms = atoms + Atoms(symbols='X16', positions=wan.centers)
view(watoms * (2, 2, 2))



# creates: citations.png
import os
import datetime

import matplotlib.pyplot as plt


months = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN',
          'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC']


def f(filename):
    papers = {}
    lines = open(filename).readlines()
    n = 0
    dois = set()
    while n < len(lines):
        line = lines[n]
        tag = line[:2]
        if tag == 'TI':
            ntitle = n
            y = None
            m = 1
            d = 15
        elif tag == 'SO':
            title = ' '.join(lines[i][3:-1] for i in range(ntitle, n))
        elif tag == 'DI':
            doi = line[3:-1]
        elif tag == 'PY':
            y = int(line.split()[1])
        elif tag == 'PD':
            for w in line.split()[1:]:
                if w[0].isdigit():
                    w = int(w)
                    if w < 100:
                        d = w
                    else:
                        y = w
                else:
                    if '-' in w:
                        w = w.split('-')[-1]
                    m = months.index(w) + 1
        elif tag == '\n' and y is not None:
            date = datetime.date(y, m, d)
            if doi not in dois:
                dois.add(doi)
                papers[doi] = (date, title)
        n += 1

    return papers


# The papers here are:
label_bib = {
    'gpaw1':
        'Mortensen et al., Phys. Rev. B (2005)',
        # http://doi.org/10.1103/PhysRevB.71.035109
    'gpaw2':
        'Enkovaara et al., J. Phys.: Condens. Matter (2010)',
        # http://doi.org/10.1088/0953-8984/22/25/253202
    'lcao':
        'Larsen et al., Phys. Rev. B (2009)',
        # http://doi.org/10.1103/PhysRevB.80.195112
    'tddft':
        'Walter et al., J. Chem. Phys. (2008)',
        # http://doi.org/10.1063/1.2943138
    'response':
        'Yan et al., Phys. Rev. B (2011)',
        # http://doi.org/10.1103/PhysRevB.83.245122
}

plt.figure(figsize=(8, 4))
total = {}
# for bib in ['gpaw1', 'tddft', 'lcao', 'gpaw2', 'response']:
for bib in ['gpaw1', 'gpaw2']:
    papers = {}
    for line in open(bib + '.txt'):
        date, doi, title = line.split(' ', 2)
        papers[doi] = (datetime.date(*[int(x) for x in date.split('-')]),
                       title.strip())
    if os.path.isfile(bib + '.bib'):
        papers.update(f(bib + '.bib'))
    papers = sorted((papers[doi][0], doi, papers[doi][1]) for doi in papers)
    plt.plot([paper[0] for paper in papers], range(1, len(papers) + 1),
             '-o', label=label_bib[bib])
    with open(bib + '.txt', 'w') as fd:
        for date, doi, title in papers:
            fd.write('%d-%02d-%02d %s %s\n' % (date.year, date.month, date.day,
                                               doi, title))
            # assert '"' not in title, title
            total[doi] = (date, title)
    x = dict([(p[1], 0) for p in papers])
    print((bib, len(papers), len(x), len(total)))


allpapers = sorted((paper[0], doi, paper[1]) for doi, paper in total.items())
plt.plot([paper[0] for paper in allpapers], range(1, len(allpapers) + 1),
         '-o', label='Total')

if 0:
    with open('citations.csv', 'w') as fd:
        n = len(allpapers)
        for date, doi, title in allpapers[::-1]:
            fd.write('%d,":doi:`%s <%s>`"\n' % (n, title, doi))
            n -= 1

plt.xlabel('date')
plt.ylabel('number of citations')
plt.legend(loc='upper left')
plt.savefig('citations.png')



# web-page: water.png
import matplotlib.pyplot as plt
import numpy as np

# Data from wm_dm_vs_scf.py
calculated_data = np.genfromtxt('water-results.txt')

# x should be number of water molecules.
# First column is number of atoms, so divide by 3 to
# obtain the number of water molecules.
x = calculated_data[:, 0] / 3

f = plt.figure(figsize=(12, 4), dpi=240)

plt.subplot(121)
plt.grid(color='k', linestyle=':', linewidth=0.3)
plt.title('Ratio of total elapsed times')
plt.ylabel(r'$T_{scf}$ / $T_{etdm}$')
plt.xlabel('Number of water molecules')
plt.ylim(1.0, 3.0)
plt.yticks(np.arange(1, 3.1, 0.5))
plt.plot(x, calculated_data[:, 1], 'bo-')

plt.subplot(122)
plt.grid(color='k', linestyle=':', linewidth=0.3)
plt.title('Ratio of elapsed times per iteration')
plt.ylabel(r'$T_{scf}$ / $T_{etdm}$')
plt.xlabel('Number of water molecules')
plt.ylim(1.0, 3.0)
plt.yticks(np.arange(1, 3.1, 0.5))
plt.plot(x, calculated_data[:, 2], 'ro-')

f.savefig("water.png", bbox_inches='tight')



# web-page: spectra_origins.png
import numpy as np
import matplotlib.pyplot as plt

fig, ax = plt.subplots(figsize=(6, 6 / 1.62))

for tag in ['COM', 'COM+x', 'COM+y', 'COM+z', '123']:
    data_ej = np.loadtxt(f'rot_spec-{tag}.dat')
    ax.plot(data_ej[:, 0], data_ej[:, 1], label=tag)

ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
ax.yaxis.set_ticks_position('left')
ax.xaxis.set_ticks_position('bottom')
plt.title(r'Rotatory strength of (R)-methyloxirane')
plt.xlabel('Energy (eV)')
plt.legend(loc='upper left', title='Origin')
plt.ylabel(r'R (10$^{-40}$ cgs eV$^{-1})$')
plt.xlim(0, 10)
plt.ylim(-80, 80)
plt.tight_layout()
plt.savefig('spectra_origins.png')



import numpy as np
from ase.build import bulk
from ase.optimize.bfgs import BFGS
from ase.constraints import UnitCellFilter
from gpaw import GPAW
from gpaw import PW

si = bulk('Si', 'fcc', a=6.0)
# Experimental Lattice constant is a=5.421 A

si.calc = GPAW(xc='PBE',
               mode=PW(400, dedecut='estimate'),
               kpts=(4, 4, 4),
               # convergence={'eigenstates': 1.e-10},  # converge tightly!
               txt='stress.txt')

uf = UnitCellFilter(si)
relax = BFGS(uf)
relax.run(fmax=0.05)  # Consider much tighter fmax!

a = np.linalg.norm(si.cell[0]) * 2**0.5
print(f'Relaxed lattice parameter: a = {a} Ang')



scalapack = False

compiler = 'icc'
libraries =['mkl_intel_lp64','mkl_sequential','mkl_cdft_core','mkl_core','pthread','m']

library_dirs = ['/nfs/slac/g/suncatfs/sw/external/intel11.1/openmpi/1.4.3/install/lib','/afs/slac/package/intel_tools/compiler11.1/mkl/lib/em64t/']

include_dirs += ['/nfs/slac/g/suncatfs/sw/external/numpy/1.4.1/install/lib64/python2.4/site-packages/numpy/core/include']

extra_link_args += ['-fPIC']

extra_compile_args = ['-I/afs/slac/package/intel_tools/compiler11.1/mkl/include','-xHOST','-O1','-ipo','-no-prec-div','-static','-std=c99','-fPIC']

define_macros =[('GPAW_NO_UNDERSCORE_CBLACS', '1'), ('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]

mpicompiler = 'mpicc'
mpilinker = mpicompiler



import numpy as np
from gpaw.new.ase_interface import GPAW
from myqueue.workflow import run


def workflow():
    with run(script='fe_sgs.py', cores=40, tmax='4h'):
        run(script='plot.py')
        run(function=check)


def check():
    energies = []
    magmoms = []
    for i in [0, 30]:
        atoms = GPAW(f'gs-{i:02}.gpw').get_atoms()
        energy = atoms.get_potential_energy()
        magmom = np.linalg.norm(atoms.calc.calculation.magmoms()[0])
        energies.append(energy)
        magmoms.append(magmom)

    print(energies, magmoms)

    assert abs(energies[0] - energies[1] - 0.0108) < 0.001
    assert abs(magmoms[0]) < 0.02
    assert abs(magmoms[1] - 0.83) < 0.02


if __name__ == '__main__':
    check()



import numpy as np
from ase.transport.calculators import TransportCalculator
import pylab

# onsite energies 0.0, nearest neighbor hopping -1.0, and
# second nearest neighbor hopping 0.2
H_lead = np.array([[0., -1., 0.2, 0.],
                   [-1., 0., -1., 0.2],
                   [0.2, -1., 0., -1.],
                   [0., 0.2, -1., 0.]])

H_scat = np.zeros((6, 6))

# Principal layers on either side of S
H_scat[:2, :2] = H_scat[-2:, -2:] = H_lead[:2, :2]

# Scattering region (hydrogen molecule) - onsite 0.0 and hopping -0.8
H_scat[2:4, 2:4] = [[0.0, -0.8], [-0.8, 0.0]]

# coupling to the leads - nearest neighbor only
H_scat[1, 2] = H_scat[2, 1] = H_scat[3, 4] = H_scat[4, 3] = 0.2

tcalc = TransportCalculator(h=H_scat,  # Scattering Hamiltonian
                            h1=H_lead,  # Lead 1 (left)
                            h2=H_lead,  # Lead 2  (right)
                            energies=np.arange(-3, 3, 0.02))

T_e = tcalc.get_transmission()
pylab.plot(tcalc.energies, T_e)
pylab.title('Transmission function')
pylab.show()

tcalc.set(pdos=[2, 3])
pdos_ne = tcalc.get_pdos()
pylab.plot(tcalc.energies, pdos_ne[0], ':')
pylab.plot(tcalc.energies, pdos_ne[1], '--')
pylab.title('Projected density of states')
pylab.show()

h_rot, s_rot, eps_n, vec_nn = tcalc.subdiagonalize_bfs([2, 3])
tcalc.set(h=h_rot, s=s_rot)  # Set the rotated matrices
for n in range(2):
    print("eigenvalue, eigenvector:", eps_n[n], ',', vec_nn[:, n])

pdos_rot_ne = tcalc.get_pdos()
pylab.plot(tcalc.energies, pdos_rot_ne[0], ':')
pylab.plot(tcalc.energies, pdos_rot_ne[1], '--')
pylab.title('Projected density of states (rotated)')
pylab.show()

h_cut, s_cut = tcalc.cutcoupling_bfs([2])
tcalc.set(h=h_cut, s=s_cut)
T_cut_bonding_e = tcalc.get_transmission()
pylab.plot(tcalc.energies, T_cut_bonding_e)
pylab.title('Transmission (bonding orbital cut)')
pylab.show()



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Catalysis: Dissociative adsorbtion of N<sub>2</sub> on a metal surface

This is the rate limiting step for ammonia synthesis.

**Scientific disclaimer:**  These calculations are done on a flat surface.  In reality, the process takes place at the foot of an atomic step on the surface.  Doing calculations on this more realistic system would be too slow for these exercises.  For the same reason, we use a metal slab with only two layers, a realistic calculation would require the double.
"""

# %%
"""
## N<sub>2</sub> Adsorption on a metal surface
This notebook shows how to calculate the adsorption energy of an N<sub>2</sub> molecule on a closepacked Ru surface. The first cell imports some modules from the ASE and GPAW packages
"""

# %%
from ase import Atoms
from gpaw import GPAW, PW
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.build import fcc111, hcp0001
import numpy as np
from ase.visualize import view
from ase.io import read, write
import time

# %%
"""
### Setting up the metal surface

Ru crystalises in the hcp structure with a lattice constants a = 2.706 Å and c = 4.282 Å.  It is often better to use the lattice constants corresponding to the DFT variant used (here PBE with PAW).  We get this from http://oqmd.org.

We model the surface by a 2 layer slab of metal atoms, and add 5Å vacuum on each side.

We visualize the system with ASE GUI, so you can check that everything looks right.  This pops up a new window.
"""

# %%
a_Ru = 2.704  # PBE value from OQMD.org; expt value is 2.706
slab = hcp0001('Ru', a=a_Ru, size=(2, 2, 2), vacuum=5.0)

# Other metals are possible, for example Rhodium
# Rhodium is FCC so you should use fcc111(...) to set up the system (same arguments).
# Remember to set the FCC lattice constant, get it from OQMD.org.

# a_Rh = 3.793
# slab = fcc111('Rh', a=a_Rh, size=(2, 2, 2), vacuum=5.0)

view(slab)

# %%
"""
To optimise the slab we need a calculator. We use the GPAW calculator in plane wave (PW) mode with the PBE exchange-correlation functional. The convergence with respect to the cutoff energy and k-point sampling should always be checked - see `Convergence.ipynb`for more information on how this can be done. For this exercise an energy cutoff of 350eV and 4x4x1 k-point mesh is chosen to give reasonable results with a limited computation time.
"""

# %%
calc = GPAW(xc='PBE',
            mode=PW(350),
            kpts={'size': (4, 4, 1), 'gamma': True},
            convergence={'eigenstates': 1e-6})
slab.calc = calc

# %%
"""
The bottom layer of the slab is fixed during optimisation. The structure is optimised until the forces on all atoms are below 0.05eV/Å.
"""

# %%
z = slab.positions[:, 2]
constraint = FixAtoms(mask=(z < z.min() + 1.0))
slab.set_constraint(constraint)
dyn = QuasiNewton(slab, trajectory='Ru.traj')
t = time.time()
dyn.run(fmax=0.05)
print(f'Wall time: {(time.time() - t) / 60} min.')

# %%
"""
The calculation will take ca. 5 minutes. While the calculation is running you can take a look at the output. How many k-points are there in total and how many are there in the irreducible part of the Brillouin zone? What does this mean for the speed of the calculation?

What are the forces and the energy after each iteration? You can read it directly in the output above, or from the saved .traj file like this:
"""

# %%
iter0 = read('Ru.traj', index=0)
print('Energy: ', iter0.get_potential_energy())
print('Forces: ', iter0.get_forces())

# %%
"""
Often you are only interested in the final energy which can be found like this:
"""

# %%
e_slab = slab.get_potential_energy()
print(e_slab)

# %%
"""
### Making a Nitrogen molecule
We now make an N<sub>2</sub> molecule and optimise it in the same unit cell as we used for the slab.
"""

# %%
d = 1.10
molecule = Atoms('2N', positions=[(0., 0., 0.), (0., 0., d)])
molecule.set_cell(slab.get_cell())
molecule.center()
calc_mol = GPAW(xc='PBE', mode=PW(350))
molecule.calc = calc_mol
dyn2 = QuasiNewton(molecule, trajectory='N2.traj')
dyn2.run(fmax=0.05)
e_N2 = molecule.get_potential_energy()

# %%
"""
We can calculate the bond length like this:
"""

# %%
d_N2 = molecule.get_distance(0, 1)
print(d_N2)

# %%
"""
How does this compare with the experimental value?
"""

# %%
"""
### Adsorbing the molecule

Now we adsorb the molecule on top of one of the Ru atoms.

Here, it would be natural to just add the molecule to the slab, and minimize.  However, that takes 45 minutes to an hour to converge, **so we cheat to speed up the calculation.**

The main slowing-down comes from the relaxation of the topmost metal atom where the N<sub>2</sub> molecule binds, this atom moves a quarter of an Ångström out.  Also, the binding length of the molecule changes when it is adsorbed, so we build a new molecule with a better starting guess.
"""

# %%
h = 1.9  # guess at the binding height
d = 1.2  # guess at the binding distance
slab.positions[4, 2] += 0.2  # pre-relax the binding metal atom.

molecule = Atoms('2N', positions=[(0, 0, 0), (0, 0, d)])
p = slab.get_positions()[4]
molecule.translate(p + (0, 0, h))
slabN2 = slab + molecule
constraint = FixAtoms(mask=(z < z.min() + 1.0))
slabN2.set_constraint(constraint)
view(slabN2)

# %%
"""
We optimise the structure.  Since we have cheated and have a good guess for the initial configuration we prevent that the optimization algorithm takes too large steps.
"""

# %%
slabN2.calc = calc
dyn = QuasiNewton(slabN2, trajectory='N2Ru-top.traj', maxstep=0.02)
t = time.time()
dyn.run(fmax=0.05)
print(f'Wall time: {(time.time() - t) / 60} min.')

# %%
"""
The calculation will take a while (10-15 minutes). While it is running please follow the guidelines in the **Exercise** section below.
"""

# %%
"""
Once the calculation is finished we can calculate the adsorption energy as:

E<sub>ads</sub> = E<sub>slab+N2</sub> - (E<sub>slab</sub> + E<sub>N2</sub>)

"""

# %%
print('Adsorption energy:', slabN2.get_potential_energy() - (e_slab + e_N2))

# %%
"""
Try to calculate the bond length of N<sub>2</sub> adsorbed on the surface. Has it changed?  What is the distance between the N<sub>2</sub> molecule and the surface?
"""

# %%
# teacher:
print('N2 bond length:', slabN2.get_distance(8, 9))


# %%


# %%
"""
### Exercise

1) Make a new notebook and set up an adsorption configuration where the N<sub>2</sub> molecule is lying down with the center of mass above a three-fold hollow site as shown below. Use an adsorption height of 1.7 Å.

<img src="N2Ru_hollow.png">

Remember that you can read in the `traj` files you have saved, so you don't need to optimise the surface again.

View the combined system before you optimize the structure to ensure that you created what you intended.
"""

# %%
slab = read('Ru.traj')
view(slab)

# %%
"""
Note that when viewing the structure, you can find the index of the individual atoms in the ``slab`` object by clicking on them.

You might also find the [`get_center_of_mass()`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html#ase.Atoms.get_center_of_mass) and [`rotate()`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html#ase.Atoms.rotate) methods useful.
"""

# %%
# teacher:
# This block is roughly what they should write themselves
h = 2.0
molecule = Atoms('2N', positions=[(0, 0, 0), (d_N2, 0, 0)])
#molecule.rotate('z',np.pi/6.,'COM')
hollow = (slab.positions[5] + slab.positions[6] + slab.positions[7]) / 3
molecule.translate(-molecule.get_center_of_mass() + hollow + (0, 0, h))
slabN2_new = slab + molecule
a = slabN2_new.repeat((2, 2, 1))
a.cell = slabN2_new.cell
write('N2Ru_hollow.png', a, show_unit_cell=1)

# %%


# %%
"""
Now you should optimize the structure as you did before with the N<sub>2</sub> molecule standing.  The calculation will probably bee too long to run interactively in a Notebook.  Prototype it here, then interrupt the calculation and copy-paste the relevant cells into a script.

Check the number of irreducible k-points and then submit the job as a batch job running on that number of CPU cores.
"""

# %%
# teacher:
calc = GPAW(xc='PBE',
            mode=PW(350),
            kpts={'size': (4, 4, 1), 'gamma': True},
            convergence={'eigenstates': 1e-6})
slabN2_new.calc = calc
dyn = QuasiNewton(slabN2_new, trajectory='N2Ru.traj', maxstep=0.02)
t = time.time()
dyn.run(fmax=0.05)
print('Wall time:', time.time() - t)

# %%


# %%
# teacher:

# Note:  Ends up with N-N of 1.287 Å and 1.65 Å above the surface

# %%
"""
3) Make a configuration where two N atoms are adsorbed in hollow sites on the surface as shown below

<img src='2NadsRu.png'>

Note that here the two N atoms sit on next-nearest hollow sites.  An alternative would be to have them on nearest neighbor sites.  If you feel energetic you could investigate that as well.  Also, there are two different kinds of hollow sites, they are not completely equivalent!
"""

# %%
# teacher:
p1 = (slab.positions[4] + slab.positions[5] + slab.positions[6])/3. + (0, 0, 2.0)
p2 = p1 + slab.positions[5] - slab.positions[4]
N1 = Atoms('N', positions=[p1])
N2 = Atoms('N', positions=[p2])
slab2Nads = slab + N1 + N2
a = slab2Nads.repeat((2, 2, 1))
a.cell = slab2Nads.cell
write('2NadsRu.png', a, show_unit_cell=1)

# %%
"""
Optimise the structure and get the final energy. Is it favourable to dissociate N<sub>2</sub> on the surface? What is the N-N distance now? What does that mean for catalysis?
"""

# %%
# teacher:
constraint = FixAtoms(mask=(z < z.min() + 1.0))
slab2Nads.set_constraint(constraint)
calc = GPAW(xc='PBE',
            mode=PW(350),
            kpts={'size': (4, 4, 1), 'gamma':True},
            convergence={'eigenstates': 1e-6})
slab2Nads.calc = calc
dyn = QuasiNewton(slab2Nads, trajectory='2Nads.traj')
dyn.run(fmax=0.05)
e_slab = slab.get_potential_energy()
e_N2 = read('N2.traj').get_potential_energy()
print('Adsorption energy:', slab2Nads.get_potential_energy() - (e_slab + e_N2))


# Generate the data visualized in web-page: g2.png
from ase.collections import g2
from doc.documentation.directmin import tools_and_data
from gpaw import LCAO, ConvergenceError
from ase.parallel import paropen

# Results (total energy, number of iterations) obtained
# in a previous calculation. Used to compare with the
# current results.
saved_results = \
    {0: tools_and_data.read_data(tools_and_data.data_g2_scf),
     1: tools_and_data.read_data(tools_and_data.data_g2_dm)}

calc_args = {'xc': 'PBE', 'h': 0.15,
             'convergence': {'density': 1.0e-6,
                             'eigenstates': 100},
             'maxiter': 333, 'basis': 'dzp',
             'mode': LCAO(), 'symmetry': 'off'}

eig_string = ['scf', 'dm']
with paropen('dm-g2-results.txt', 'w') as fdm, \
        paropen('scf-g2-results.txt', 'w') as fscf:
    fd = {0: fscf, 1: fdm}
    for name in saved_results[0].keys():
        atoms = g2[name]
        atoms.center(vacuum=7.0)
        for dm in [0, 1]:
            txt = name + eig_string[dm] + '.txt'
            tools_and_data.set_calc(atoms, calc_args, txt, dm)

            try:
                e, iters, t = tools_and_data.get_energy_and_iters(atoms, dm)

                # Compare with saved results from previous calculation
                e_diff_saved_calc = abs(saved_results[dm][name][1] - e)
                iters_diff_saved_calc = abs(saved_results[dm][name][0] - iters)
                tools_and_data.compare_calculated_and_saved_results(
                    e_diff_saved_calc, iters_diff_saved_calc,
                    eig_string, name, dm)

                print(name + "\t{}".format(iters),
                      file=fd[dm], flush=True)

            except ConvergenceError:
                print(name + "\t{}".format(None),
                      file=fd[dm], flush=True)



# Generate the data visualized in web-page: water.png
import numpy as np
from doc.documentation.directmin import tools_and_data
from ase import Atoms
from gpaw import LCAO, ConvergenceError
from ase.parallel import paropen
from gpaw.mpi import world
from gpaw.atom.basis import BasisMaker
from gpaw import setup_paths
setup_paths.insert(0, '.')

for symbol in ['H', 'O']:
    bm = BasisMaker(symbol, xc='PBE')
    basis = bm.generate(zetacount=3, polarizationcount=2)
    basis.write_xml()

positions = tools_and_data.positions

L = 9.8553729
r = [[1, 1, 1],
     [2, 1, 1],
     [2, 2, 1]]
calc_args = {'xc': 'PBE', 'h': 0.2,
             'convergence': {'density': 1.0e-6,
                             'eigenstates': 100},
             'maxiter': 333, 'basis': 'tzdp',
             'mode': LCAO(), 'symmetry': 'off',
             'parallel': {'domain': world.size}}
# Results (total energy, number of iterations) obtained
# in a previous calculation. Used to compare with the
# current results.
saved_results = tools_and_data.wm_saved_results

eig_string = ['scf', 'dm']
t = np.zeros(2)
iters = np.zeros(2)
with paropen('water-results.txt', 'w') as fd:
    for i, x in enumerate(r):
        atoms = Atoms('32(OH2)', positions=positions)
        atoms.set_cell((L, L, L))
        atoms.set_pbc(1)
        atoms = atoms.repeat(x)
        name = str(len(atoms) // 3) + '_H2Omlcls'
        try:
            for dm in [0, 1]:
                txt = name + '_' + eig_string[dm] + '.txt'
                tools_and_data.set_calc(atoms, calc_args, txt, dm)

                e, iters[dm], t[dm] = \
                    tools_and_data.get_energy_and_iters(atoms, dm)

                # Compare with saved results from previous calculation
                e_diff_saved_calc = abs(saved_results[dm][i, 0] - e)
                iters_diff_saved_calc = \
                    abs(saved_results[dm][i, 1] - iters[dm])
                tools_and_data.compare_calculated_and_saved_results(
                    e_diff_saved_calc, iters_diff_saved_calc,
                    eig_string, name, dm)

            # Ratio of elapsed times per iteration
            # 2 is added to account for the diagonalization
            # performed at the beginning and at the end of etdm
            ratio_per_iter = (t[0] / iters[0]) / (t[1] / (iters[1] + 2))

            print("{}\t{}\t{}".format(
                  len(atoms), t[0] / t[1], ratio_per_iter),
                  flush=True, file=fd)

        except ConvergenceError:
            print("{}\t{}\t{}".format(
                len(atoms), None, None),
                flush=True, file=fd)



def link_shared_object (self,
                     objects,
                     output_filename,
                     output_dir=None,
                     libraries=None,
                     library_dirs=None,
                     runtime_library_dirs=None,
                     export_symbols=None,
                     debug=0,
                     extra_preargs=None,
                     extra_postargs=None,
                     build_temp=None,
                     target_lang=None):

    if output_dir is None:
        (output_dir, output_filename) = os.path.split(output_filename)
    output_fullname = os.path.join(output_dir, output_filename)
    output_fullname = os.path.abspath(output_fullname)
    linkline = f"{output_filename[:-2]} {output_fullname}"
    for l in library_dirs:
        linkline += " -L" + l
    for l in libraries:
        linkline += " -l" + l
    old_fmt = self.static_lib_format
    self.static_lib_format = "%s%.0s"
    self.create_static_lib(objects,
                           output_filename,
                           output_dir,
                           debug,
                           target_lang)

    self.static_lib_format = old_fmt
    print("Append to Setup: ", linkline)



import traceback
from time import time
from glob import glob

import ase.db

import gpaw.mpi
from gpaw import GPAW

exec(open('params.py').read())  # get the calc() function
setup = globals()['calc']

c = ase.db.connect('results.db')

names = [name for name in glob('*.py') if name not in
         ['run.py', 'params.py', 'submit.agts.py', 'analyse.py']]

for name in names:
    namespace = {}
    exec(open(name).read(), namespace)
    atoms = namespace['atoms']
    ncpus = namespace.get('ncpus', 8)

    if gpaw.mpi.size != ncpus:
        continue

    name = name[:-3]

    id = c.reserve(name=name)
    if not id:
        continue

    if atoms.calc is None:
        atoms.calc = GPAW()

    setup(atoms)

    atoms.calc.set(txt=name + '.txt')

    t = time()
    try:
        e1 = atoms.get_potential_energy()
        ok = True
    except Exception:
        ok = False
        if gpaw.mpi.rank == 0:
            traceback.print_exc(file=open(name + '.error', 'w'))
    t = time() - t

    c.write(atoms, name=name, ok=ok,
            time=t, iters=atoms.calc.get_number_of_iterations(), ncpus=ncpus)

    del c[id]



from ase.build import bulk
from gpaw import GPAW

ag = bulk('Ag')
calc = GPAW(mode='pw',
            xc='LDA',
            kpts=(10, 10, 10),
            txt='Ag_LDA.txt')
ag.calc = calc
ag.get_potential_energy()
calc.write('Ag_LDA.gpw')
with open('eels.py') as fd:
    code = fd.read()
exec(code)



"""Calculate PBE0 bandgaps.

The numbers are compared to:

  Hybrid functionals within the all-electron FLAPW method:
  implementation and applications of PBE0

  Markus Betzinger, Christoph Friedrich, Stefan Blügel

  Physical Review B 81, 195117 (2010)
  DOI: 10.1103/PhysRevB.81.195117
"""

import numpy as np
import ase.db
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack

from gpaw import GPAW, PW
from gpaw.hybrids.eigenvalues import non_self_consistent_eigenvalues as nsceigs


# Data from Betzinger, Friedrich and Blügel:
bfb = {
    'Si': ['diamond', 5.421,
           2.56, 3.96, 2.57, 3.97,
           0.71, 1.93, 0.71, 1.93,
           1.54, 2.87, 1.54, 2.88],
    'C': ['diamond', 3.553,
          5.64, 7.74, 5.59, 7.69,
          4.79, 6.69, 4.76, 6.66,
          8.58, 10.88, 8.46, 10.77],
    'GaAs': ['zincblende', 5.640,
             0.55, 2.02, 0.56, 2.01,
             1.47, 2.69, 1.46, 2.67,
             1.02, 2.38, 1.02, 2.37],
    'MgO': ['rocksalt', 4.189,
            4.84, 7.31, 4.75, 7.24,
            9.15, 11.63, 9.15, 11.67,
            8.01, 10.51, 7.91, 10.38],
    'NaCl': ['rocksalt', 5.569,
             5.08, 7.13, 5.2, 7.26,
             7.39, 9.59, 7.6, 9.66,
             7.29, 9.33, 7.32, 9.41],
    'Ar': ['fcc', 5.26,
           8.71, 11.15, 8.68, 11.09]}

nk = 12
kpts = monkhorst_pack((nk, nk, nk)) + 0.5 / nk

c = ase.db.connect('gaps.db')

for name in ['Si', 'C', 'GaAs', 'MgO', 'NaCl', 'Ar']:
    id = c.reserve(name=name)
    if id is None:
        continue

    x, a = bfb[name][:2]
    atoms = bulk(name, x, a=a)
    atoms.calc = GPAW(xc='PBE',
                      mode=PW(500),
                      parallel=dict(band=1),
                      nbands=-8,
                      convergence=dict(bands=-7),
                      kpts=kpts,
                      txt='%s.txt' % name)

    atoms.get_potential_energy()
    atoms.calc.write(name + '.gpw', mode='all')

    ibzk_kc = atoms.calc.get_ibz_k_points()
    n = int(atoms.calc.get_number_of_electrons()) // 2

    ibzk = []
    for k_c in [(0, 0, 0), (0.5, 0.5, 0), (0.5, 0.5, 0.5)]:
        k = abs(ibzk_kc - k_c).max(1).argmin()
        ibzk.append(k)
        if name == 'Ar':
            break

    eps_skn, deps_skn, deps0_skn = nsceigs(name + '.gpw',
                                           'PBE0',
                                           n - 1, n + 1,
                                           ibzk)

    eps0_kn = (eps_skn - deps_skn + deps0_skn)[0]

    data = {}
    for k, point in enumerate('GXL'):
        data[point] = np.array(
            [eps_skn[0, k, 1] - eps_skn[0, 0, 0],
             eps0_kn[k, 1] - eps0_kn[0, 0]] +
            bfb[name][2 + k * 4:6 + k * 4])
        if name == 'Ar':
            break

    c.write(atoms, name=name, data=data)
    del c[id]



# web-page: phases.png
import numpy as np
import pylab as plt

a = np.load('phases_7x200.npz')
phit_km = a['phi_km']
St_km = a['S_km']
Nk = len(phit_km)

phi_km = np.zeros((len(phit_km) + 1, len(phit_km[0])), float)
phi_km[1:] = phit_km
phi_km[0] = phit_km[-1]
S_km = np.zeros((len(phit_km) + 1, len(phit_km[0])), float)
S_km[1:] = St_km
S_km[0] = St_km[-1]
# S_km += 1
S_km /= 2

Nm = len(phi_km[0])
phi_km = np.tile(phi_km, (1, 2))
phi_km[:, Nm:] += 2 * np.pi
S_km = np.tile(S_km, (1, 2))

plt.figure()
plt.scatter(np.tile(np.arange(len(phi_km)), len(phi_km.T)),
            phi_km.T.reshape(-1),
            cmap=plt.get_cmap('viridis'),
            c=S_km.T.reshape(-1),
            s=5,
            marker='o')

cbar = plt.colorbar()
cbar.set_label(r'$\langle S_z\rangle/\hbar$', size=20)

# plt.xlabel(r'$k_\mathrm{y}$', size=24)
plt.ylabel(r'$\gamma_x$', size=24)
plt.xticks([0, Nk / 2, Nk],
           [r'$-\mathrm{M}$', r'$\Gamma$', r'$\mathrm{M}$'], size=20)
plt.yticks([0, np.pi, 2 * np.pi], [r'$0$', r'$\pi$', r'$2\pi$'], size=20)
plt.axis([0, Nk, 0, 2 * np.pi])
plt.tight_layout()
plt.savefig('phases.png')
# plt.show()
